"""Unit-tests for the physics module.

Run as python test_post_process.py in console or via coverage
"""
from __future__ import annotations
import unittest

import longitudinal_tomography.data.data_treatment as treat
import longitudinal_tomography.data.post_process as post_process
from .. import commons

# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestPostProcess(unittest.TestCase):
    def test_emittance_rms(self):
        tomo, machine = commons.get_tomography_params()

        t_bins, e_bins, phase_space = treat.phase_space(tomo, machine)

        emittance_rms = post_process.emittance_rms(phase_space, t_bins, e_bins)

        self.assertAlmostEqual(emittance_rms, 0.06802901508017638)

    def test_emittance_90(self):
        tomo, machine = commons.get_tomography_params()

        t_bins, e_bins, phase_space = treat.phase_space(tomo, machine)

        emittance_90 = post_process.emittance_90(phase_space, t_bins, e_bins)

        self.assertAlmostEqual(emittance_90, 0.269939386246181)

    def test_emittance_fractional(self):
        tomo, machine = commons.get_tomography_params()

        t_bins, e_bins, phase_space = treat.phase_space(tomo, machine)

        emittance = post_process.emittance_fractional(phase_space, t_bins,
                                                      e_bins, fraction=80)

        self.assertAlmostEqual(emittance, 0.2183279430331099)

    def test_emittance_fractional_bounds(self):
        tomo, machine = commons.get_tomography_params()

        t_bins, e_bins, phase_space = treat.phase_space(tomo, machine)

        with self.assertRaises(ValueError):
            post_process.emittance_fractional(phase_space, t_bins,
                                              e_bins, fraction=-1)

            post_process.emittance_fractional(phase_space, t_bins,
                                              e_bins, fraction=101)

    def test_rms_dpp(self):
        tomo, machine = commons.get_tomography_params()

        energy = 1098272089.0462158
        momentum = 570830158.7660657
        mass = 938272088.1604904

        t_bins, e_bins, phase_space = treat.phase_space(tomo, machine)

        rms_dpp = post_process.rms_dpp(phase_space.sum(0), e_bins,
                                       energy, mass)
        self.assertAlmostEqual(rms_dpp, 0.0005028782274293471)

    def test_post_process(self):
        tomo, machine = commons.get_tomography_params()

        energy = 1098272089.0462158
        momentum = 570830158.7660657
        mass = 938272088.1604904

        t_bins, e_bins, phase_space = treat.phase_space(tomo, machine)
        processed_values = post_process.post_process(phase_space, t_bins,
                                                     e_bins, energy, mass)

        correct = {
            'emittance_rms': 0.06802901508017638,
            'emittance_90': 0.269939386246181,
            'rms_dp/p': 0.0005028782274293471,
        }

        for k, v in correct.items():
            self.assertAlmostEqual(v, processed_values[k])



"""Module containing functions for treatment of data.

:Author(s): **Christoffer Hjertø Grindheim**, **Anton Lu**
"""
from __future__ import annotations
from typing import TYPE_CHECKING
from warnings import warn

import numpy as np
import itertools as itl

from .. import exceptions as expt
from ..utils import physics
from ..utils import tomo_config as conf
from ..cpp_routines import libtomo
from ..python_routines import data_treatment
from . import pre_process

import logging

if TYPE_CHECKING:
    from .profiles import Profiles
    from ..tracking.machine import Machine
    from ..tracking.machine_base import MachineABC
    from ..tomography.__tomography import TomographyABC

    from typing import Iterable, Union, Tuple

    FloatArr = np.ndarray[float]

__all__ = ['rebin', 'fit_synch_part_x', 'phase_space']

log = logging.getLogger(__name__)


def rebin(waterfall: Iterable[float], rbn: int, dtbin: float = None,
          synch_part_x: float = None) \
        -> Union[Tuple[FloatArr, float, float], Tuple[FloatArr, float]]:
    """
    Rebin waterfall from shape (P, X) to (P, Y).
    P is the number of profiles, X is the original number of bins,
    and Y is the number of bins after the re-binning.

    The algorithm is based on the rebinning function from the
    original tomography program.

    An array of length N, rebinned with a rebin factor of R will result
    in a array of length N / R. If N is not dividable on R, the resulting
    array will have the length (N / R) + 1.

    Parameters
    ----------
    waterfall: Iterable[float]
        2D array of raw-data shaped as waterfall. Shape: (nprofiles, nbins).
    rbn: int
        Rebinning factor.
    dtbin: float
        Size of profile bins [s]. If provided, the function
        will return the new size of the bins after rebinning.
    synch_part_x: float
        x-coordinate of synchronous particle, measured in bins.
        If provided, the function will return its updated coordinate.

    Returns
    -------
    rebinned: ndarray
        Rebinned waterfall
    dtbin: float, optional, default=None
        If a dtbin has been provided in the arguments, the
        new size of the profile bins will be returned. Otherwise,
        None will be returned.
    synch_part_x: float, optional, default=None
        If a synch_part_x has been provided in the arguments, the
        new x-coordinate of the synchronous particle in bins will be returned.
        Otherwise, None will be returned.
    """
    # warn('The rebin function has been moved to '
    #      'longitudinal_tomography.data.pre_process')
    return pre_process.rebin(waterfall, rbn, dtbin, synch_part_x)


# Original function for finding synch_part_x
# Finds synch_part_x based on a linear fit on a reference profile.
def fit_synch_part_x(profiles: Profiles) -> Tuple[FloatArr, float, float]:
    """
    Linear fit to estimate the phase coordinate of the synchronous
    particle. The found phase is returned as a x-coordinate of the phase space
    coordinate systems in fractions of bins. The estimation is done at
    the beam reference profile, which is set in the
    :class:`longitudinal_tomography.tracking.machine.Machine` object.

    Parameters
    ----------
    profiles: Profiles
        Profiles object containing waterfall and information about the
        measurement.

    Returns
    -------
    fitted_synch_part_x
        X coordinate in the phase space coordinate system of the synchronous
        particle given in bin numbers.
    lower bunch limit
        Estimation of the lower bunch limit in bin numbers.
        Needed for
        :func:`longitudinal_tomography.utils.tomo_output.write_plotinfo_ftn`
        function in order to write the original output format.
    upper bunch limit
        Estimation of the upper bunch limit in bin numbers.
        Needed for
        :func:`longitudinal_tomography.utils.tomo_output.write_plotinfo_ftn`
        function in order to write the original output format.
    """
    # warn('The fit_synch_part_x function has moved to '
    #      'longitudinal_tomography.data.pre_process')
    return pre_process.fit_synch_part_x(profiles)


def phase_space(tomo: TomographyABC, machine: MachineABC,
                reconstr_idx: int = 0) \
        -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Returns time, energy and phase space density arrays from a
    reconstruction, requires the homogenous distribution to have been
    generated by the particles class.

    Parameters
    ----------
    tomo: TomographyABC
        Object holding the information about a tomographic reconstruction.
    machine: Machine
        Object holding information about machine and reconstruction parameters.
    reconstr_idx: int
        Index of profile to be reconstructed.

    Returns
    -------
    t_range: ndarray
        1D array containing time axis of reconstructed phase space image.
    E_range: ndarray
        1D array containing energy axis of reconstructed phase space image.
    density: ndarray
        2D array containing the reconstructed phase space image.

    Raises
    ------
    InputError: Exception
        phase_space function requires automatic phase space generation
        to have been used.
    """
    if machine.dEbin is None:
        raise expt.InputError("""phase_space function requires automatic
                              phase space generation to have been used""")

    return phase_space_from_coordinates(tomo.xp[:, reconstr_idx],
                                        tomo.yp[:, reconstr_idx],
                                        tomo.weight, machine.nbins,
                                        machine.synch_part_x,
                                        machine.synch_part_y, machine.dtbin,
                                        machine.dEbin)


def phase_space_from_coordinates(xp: Iterable[int], yp: Iterable[int],
                                 weight: Iterable[int], nbins: int,
                                 synch_x: float, synch_y: float,
                                 dtbin: float, dEbin: float) \
                                        -> Tuple[FloatArr, FloatArr, FloatArr]:
    """
    Returns time, energy and phase space density arrays from a
    reconstruction.

    Parameters
    ----------
    xp : np.ndarray
        An array of x-coordinates in bin numbers for the reconstructed phase
        space.
    yp : np.ndarray
        An array of y-coordinates in bin numbers for the reconstructed phase
        space.
    weight : np.ndarray
        The array of particle weights for the reconstructed phase space.
    nbins : int
        Number of bins in each dimension.
    synch_x : float
        The bin number in time of the synchronous particle.
    synch_y : float
        The bin number in energy of the synchronous particle.
    dtbin : float
        The number of s per time bin.
    dEbin : float
        The number of eV per energy bin.

    Returns
    -------
    t_range : np.ndarray
        1D array of the center times for each time bin.
    E_range : np.ndarray
        1D array of the center energies for each energy bin.
    density : np.ndarray
        2D array of bin weights in phase space.
    """

    density = conf.make_phase_space(xp, yp, weight, nbins)

    t_range = (np.arange(nbins) - synch_x) * dtbin
    E_range = (np.arange(nbins) - synch_y) * dEbin

    return t_range, E_range, density


# Creates a [nbins, nbins] array populated with the weights of each test
# particle
def _make_phase_space(xp: Iterable[float], yp: Iterable[float],
                      weights: Iterable[float], nbins: int) -> FloatArr:
    log.warning('tomo.data.data_treatment._make_phase_space has moved to '
                'the C++ library at '
                'tomo.cpp_routines.libtomo.make_phase_space. '
                'This function is provided for backwards compatibility '
                'and can be removed without further notice.')
    return libtomo.make_phase_space(xp, yp, weights, nbins)


def calc_baseline_ftn(*args):
    # warn('This function has moved to '
    #      'longitudinal_tomography.compat.fortran.')
    from longitudinal_tomography.compat.fortran import calc_baseline
    return calc_baseline(*args)


def density_to_macro(tRange: Iterable[float], ERange: Iterable[float],
                     density: Iterable[float], n_macro: int,
                     threshold: float = 1E-5) -> Tuple[FloatArr, FloatArr]:
    """
    Takes a time range, energy range and density function and converts it to
    n_macro number macroparticles to reproduce the distribution via randomly
    sampling from it.


    Parameters
    ----------
    t_range : np.ndarray
        1D array of the center times for each time bin.
    E_range : np.ndarray
        1D array of the center energies for each energy bin.
    density : np.ndarray
        2D array of bin weights in phase space.
    n_macro: int
        Amount of macroparticles to be generated
    threshold: float
        Threshhold under which density is set to 0.

    Returns
    -------
    dt: np.ndarray
        1D array of time coordinates of each macroparticle
    dE: np.ndarray
        1D array of energy coordinates of each macroparticle
    """

    dtbin = tRange[1] - tRange[0]
    dEbin = ERange[1] - ERange[0]

    sumDens = np.sum(density)
    density[density / sumDens < threshold] = 0

    prob = density.flatten()
    prob = prob / np.sum(prob)
    coords = np.array(list(itl.product(tRange, ERange)))
    n_macro = int(n_macro)

    samples = np.random.choice(prob.shape[0], n_macro, p=prob)

    dt = coords[samples, 0] + dtbin * (np.random.random(n_macro) - 0.5)
    dE = coords[samples, 1] + dEbin * (np.random.random(n_macro) - 0.5)

    return dt, dE



class TomoException(Exception):
    pass

# ==========================
# CUDA Compilation Exception
# ==========================

class CudaCompilationException(Exception):
    pass

# ================
# Array Exceptions
# ================


class UnequalArrayShapes(TomoException, ValueError):
    pass

# ====================
# Numerical Exceptions
# ====================


class OverMaxIterationsError(TomoException):
    pass


class ValuesOutOfBrackets(TomoException):
    pass


class ArrayLengthError(TomoException, ValueError):
    pass


class ArrayElementsNotEqualError(TomoException, ValueError):
    pass

# ===========================
# Input Parameters Exceptions
# ===========================


class InputError(TomoException):
    pass


class MachineParameterError(TomoException, ValueError):
    pass


class SpaceChargeParameterError(TomoException, ValueError):
    pass


class RawDataImportError(TomoException):
    pass


class ArgumentError(TomoException, ValueError):
    pass


class NegativeIndexError(IndexError):
    pass

# ===========================
# Time space Exceptions
# ===========================


class RebinningError(TomoException, RuntimeError):
    pass


class WaterfallReducedToZero(TomoException, RuntimeError):
    pass


class FilteredProfilesError(TomoException, ValueError):
    pass


class WaterfallError(TomoException, ValueError):
    pass


class ProfileChargeNotCalculated(TomoException, ValueError):
    pass

# ============================
# MapInfo exceptions
# ============================


class EnergyBinningError(TomoException, ValueError):
    pass


class EnergyLimitsError(TomoException, ValueError):
    pass


class PhaseLimitsError(TomoException, ValueError):
    pass


class MapCreationError(TomoException):
    pass


# ============================
# PARTICLE TRACKING EXCEPTIONS
# ============================

class InvalidParticleError(TomoException, ValueError):
    pass


class TrackingError(TomoException, RuntimeError):
    pass


class SelfFieldTrackingError(TomoException):
    pass

# ============================
# TOMOGRAPHY EXCEPTIONS
# ============================


class PhaseSpaceReducedToZeroes(TomoException, RuntimeError):
    pass


class XPOutOfImageWidthError(TomoException, ValueError):
    pass


class CoordinateImportError(TomoException, ValueError):
    pass


class CoordinateError(TomoException, ValueError):
    pass

# ============================
# ANALYSIS EXCEPTIONS
# ============================

class InvalidProfileError(TomoException, ValueError):
    pass



# coding: utf-8
'''
tomographics.m translated in py
Work in progress
A. Lasheen

TODO: include double RF
TODO: documentation
TODO: include plotting routine matching the mathematica output (draft already exists)
TODO: docstrings
TODO: unit tests
'''
from __future__ import annotations

# General imports
import scipy.signal
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from typing import TYPE_CHECKING

# Local imports
from .. import exceptions as exc

if TYPE_CHECKING:
    from typing import Iterable, Tuple
    from ..tracking.machine_base import MachineABC

    FloatArr = np.ndarray[float]
    IntArray = np.ndarray[int]


def foot_tangent_fit_dq(x: Iterable[float], y: Iterable[float],
                        t_rf: float, apply_filter: bool=True)\
                                                        -> Tuple[float, float]:
    """
    Compute the x-intercepts of the foot tangent fit

    Parameters
    ----------
    x : Iterable[float]
        The x-coordinates of the profile.
    y : Iterable[float]
        The y-coordinates of the profile.
    t_rf : float
        The RF period.
    apply_filter : bool, optional
        DESCRIPTION. The default is True.

    Raises
    ------
    InvalidProfileError
        If the profile is too long for a valid fit, or the resulting fit
        goes past an RF period, an InvalidProfileError is raised.

    Returns
    -------
    Tuple[float, float]
        The x coordinates of the lower and upper intercepts.
    """
    if apply_filter:
        y = scipy.signal.savgol_filter(y, 5, 4)

    # Threshold at 15% of the peak
    threshold = 0.15 * np.max(y)
    indices_above = np.where(y >= threshold)[0]

    # Find the indices corresponding to 15%
    left_index = indices_above[0]
    right_index = indices_above[-1]

    if (left_index-3 < 0) or (right_index+3 >= len(x)):
        raise exc.InvalidProfileError("The input profile is too long for a "
                                      +"foot tangent fit ")

    # Polynomial fits of order 1 (straight lines) to find the two
    # tangent lines.  For each fit, 6 points of the y are used, 3 to the
    # left of the 15% point and 3 to the right of the 15% point.
    coefficient_1 = np.polyfit(x[left_index-3:left_index+3],
                               y[left_index-3:left_index+3], 1)
    coefficient_2 = np.polyfit(x[right_index-2:right_index+4],
                               y[right_index-2:right_index+4], 1)

    # Find the intersections of the two tangents with the baseline, here
    # supposed at 0.
    x_min = - coefficient_1[1] / coefficient_1[0]
    x_max = - coefficient_2[1] / coefficient_2[0]

    # Obvious checks, but just to be sure...
    if (0 <= (x_max - x_min) < t_rf):
        return [x_min, x_max]
    else:
        raise exc.InvalidProfileError("The computed x-intercepts go beyond"
                                      +" one RF period")


def foot_tangent_fit(y: Iterable[float], dx: float = 1) -> Tuple[float, float]:
    """
    Compute the foot tangent fit

    Parameters
    ----------
    y : Iterable[float]
        The bunch profile to be fitted.
    dx : float, optional
        The bin width of the profile. The default is 1.

    Returns
    -------
    Tuple[float, float]
        The positions of the lower and upper points of the fit in
        units of dx.
    """

    threshold = 0.15 * np.max(y)
    max_index = np.where(y == np.max(y))[0][0]
    indices_below = np.where(y < threshold)[0]

    left_index = indices_below[indices_below < max_index][-1]
    right_index = indices_below[indices_below > max_index][0]

    y_left = y[left_index - 1:left_index + 3]
    y_right = y[right_index - 2:right_index + 2]

    foot_left = left_index - 5*np.sum(y_left)/(-6*y_left[0] - 2*y_left[1]
                                               +2*y_left[2] + 6*y_left[3])
    foot_right = right_index - 6*np.sum(y_right)/(-6*y_right[0] - 2*y_right[1]
                                                  +2*y_right[2] + 6*y_right[3])

    return foot_left * dx, foot_right * dx


# TODO: What's this for?
def foot_tangent_fit_density(y: Iterable[float], x: Iterable[float],
                             level: float = 0.15) -> float:

    if not (0 < level < 1):
        raise ValueError("Level must be between 0 and 1")

    y_zero = y-np.min(y)

    threshold = level * np.max(y_zero)
    max_index = np.where(y_zero == np.max(y_zero))[0][0]
    indices_below = np.where(y_zero < threshold)[0]

    right_index = indices_below[indices_below > max_index][0]

    y_right = y_zero[right_index - 2:right_index + 2]

    foot_right = right_index - 6*np.sum(y_right)/(-6*y_right[0] - 2*y_right[1]
                                                  +2*y_right[2] + 6*y_right[3])

    foot_density = np.interp(foot_right, np.arange(len(x)), x)

    return foot_density


def _urf_length_at_level(phi_array: Iterable[float], urf: Iterable[float],
                         phi_0: float, urf_level: float)\
                                                 -> Tuple[float, float, float]:

    urf_left = urf[phi_array < phi_0] - urf_level
    urf_right = urf[phi_array > phi_0] - urf_level

    phi_left = np.interp(0, urf_left, phi_array[phi_array < phi_0])
    phi_right = np.interp(0, urf_right[np.argsort(urf_right)],
                          phi_array[phi_array > phi_0][np.argsort(urf_right)])
    phi_tot = phi_right - phi_left

    return phi_tot, phi_right, phi_left


def _residue_urf_length(urf_level: float,
                        phi_target: float, phi_array: float,
                        urf: float, phi_0: float) -> float:

    phi_tot = _urf_length_at_level(phi_array, urf, phi_0, urf_level)[0]
    residue = (phi_tot - phi_target)**2.

    return residue


def matched_area_calc(tomomachine: MachineABC, bunch_length: float,
                      idx_frame: int=None) -> float:
    """
    Compute the matched area for a given bunch length at the specified
    profile number.

    Parameters
    ----------
    tomomachine : MachineABC
        The machine object used for the reconstruction.
    bunch_length : float
        #TODO: Confirm length unit
        The bunch length in seconds.
    idx_frame : int, optional
        The profile number to compute the emittance at.
        The default is None.
        If None, the machine reference frame is used.

    Returns
    -------
    float
        The computed matched area in eVs.
    """

    if idx_frame is None:
        idx_frame = tomomachine.machine_ref_frame

    turn = int(idx_frame * tomomachine.dturns)

    phi_0 = tomomachine.phi0[turn]
    energy = tomomachine.e0[turn]
    eta0 = tomomachine.eta0[turn]
    omega_rev0 = tomomachine.omega_rev0[turn]

    gamma0 = energy / tomomachine.e_rest
    bunch_length_phase = bunch_length * omega_rev0 * tomomachine.h_num

    phi_array = phi_0 + np.linspace(-np.pi, np.pi, 1000)
    urf = (tomomachine.vrf1 * (np.cos(phi_0) - np.cos(phi_array))
           + (phi_0 - phi_array) * tomomachine.vrf1 * np.sin(phi_0))

    urf_level_opti = minimize(_residue_urf_length,
                              (np.max(urf) - np.min(urf))/2 + np.min(urf),
                              args=(bunch_length_phase, phi_array, urf, phi_0),
                              method='Powell')['x']

    phi_tot, phi_right, phi_left = _urf_length_at_level(phi_array, urf, phi_0,
                                                        urf_level_opti)
    phi_array_matched = np.linspace(phi_left, phi_right, 1000)

    dE_height_array = np.abs(np.sqrt(tomomachine.q*(gamma0 - 1/gamma0)
                                     * tomomachine.e_rest/(np.pi
                                                           * tomomachine.h_num
                                                           * eta0)
                                     * (tomomachine.vrf1
                                        * (np.cos(phi_array_matched)
                                           - np.cos(phi_left))
                                        + (phi_array_matched - phi_left)
                                        * tomomachine.vrf1 * np.sin(phi_0))
                                     )
                             )

    dE_height_array[np.isnan(dE_height_array)] = 0

    matched_area = (2 / (tomomachine.h_num*omega_rev0)
                    * np.trapz(dE_height_array,
                               dx=phi_array_matched[1] - phi_array_matched[0]))

    return matched_area


def _cumulative_density_calc(tomo_image: Iterable[float], dt: float,
                             dE: float) -> Tuple[np.ndarray]:

    cumulative_density = np.cumsum(-np.sort(-tomo_image.flatten()))
    cumulative_density_x_array = np.arange(len(cumulative_density)) * dt * dE

    return cumulative_density, cumulative_density_x_array


def emittance_density_calc(tomo_image: Iterable[float], dt: float,
                           dE: float, density_target: float=0.9) -> float:
    """
    Compute the emittance that contains a specified fraction of the beam
    density.

    Parameters
    ----------
    tomo_image : Iterable[float]
        The 2-array defining the phase space distribution.
    # TODO: Confirm units
    dt : float
        The time in seconds of one bin.
    dE : float
        The energy in eV of one bin.
    density_target : float, optional
        The fraction of beam density to compute the emittance for.
        The default is 0.9.

    Raises
    ------
    ValueError
        If density_target is not between 0 and 1, a ValueError is raised.

    Returns
    -------
    float
        The computed emittance in eVs.
    """

    if not (0 < density_target < 1):
        raise ValueError("density_target must be between 0 and 1")

    cumulative_density, cumulative_density_x_array = _cumulative_density_calc(
                                                            tomo_image, dt, dE)

    emittance = np.interp(density_target, cumulative_density,
                          cumulative_density_x_array)

    return emittance


def rms_params(tomo_image: Iterable[float], dt: float, dE: float)\
                                   -> Tuple[float, float, float, float, float]:
    """
    Compute the RMS emittance, mean dt, RMS dt, mean dE and RMS dE

    Parameters
    ----------
    tomo_image : Iterable[float]
        The 2-array containing the phase space distribution.
    # TODO: Confirm units
    dt : float
        The time in seconds of one bin.
    dE : float
        The energy in eV of one bin.

    Returns
    -------
    Tuple[float, float, float, float, float]
        RMS emittance, mean dt, RMS dt, mean dE and RMS dE.
    """

    y_matrix_tomo1, x_matrix_tomo1 = np.meshgrid(np.arange(tomo_image.shape[0]),
                                                 np.arange(tomo_image.shape[1]))
    xbar = np.sum(tomo_image * x_matrix_tomo1)
    xms = np.sum(tomo_image * x_matrix_tomo1**2.)
    ybar = np.sum(tomo_image * y_matrix_tomo1)
    yms = np.sum(tomo_image * y_matrix_tomo1**2.)
    xybar = np.sum(tomo_image * x_matrix_tomo1 * y_matrix_tomo1)

    rmsemittance = np.pi*dt*dE*np.sqrt((xms - xbar**2.)
                                       * (yms - ybar**2.)
                                       - (xybar - xbar*ybar)**2.)

    mean_dt = xbar * dt
    sigma_dt = np.sqrt(xms - xbar**2.) * dt

    mean_dE = (ybar - tomo_image.shape[0]/2) * dE
    sigma_dE = np.sqrt(yms - ybar**2.) * dE

    return rmsemittance, mean_dt, sigma_dt, mean_dE, sigma_dE


def density_vs_emittance(tomomachine: MachineABC, tomo_image: Iterable[float],
                         time_array: Iterable[float], dE_array: Iterable[float],
                         n_points_amplitude: int=100, idx_frame: int=None)\
                                   -> Tuple[float, float, float, float, float]:
    """
    TODO: Good description?

    Parameters
    ----------
    tomomachine : MachineABC
        The machine object used for the reconstruction.
    tomo_image : Iterable[float]
        The 2-array defining the phase space distribution.
    time_array : Iterable[float]
        The time axis of the phase space.
    dE_array : Iterable[float]
        The energy axis of the phase space.
    n_points_amplitude : int, optional
        TODO: DESCRIPTION. The default is 100.
    idx_frame : int, optional
        The profile number to compute at.
        The default is None.
        If None, the machine reference frame is used.

    Returns
    -------
    Tuple[float, float, float, float, float]
        TODO: DESCRIPTION.
    """

    if idx_frame is None:
        idx_frame = tomomachine.machine_ref_frame

    turn = int(idx_frame * tomomachine.dturns)

    phi_0 = tomomachine.phi0[turn]
    energy = tomomachine.e0[turn]
    eta0 = tomomachine.eta0[turn]
    h_num = tomomachine.h_num
    omega_rev0 = tomomachine.omega_rev0[turn]
    gamma0 = energy / tomomachine.e_rest

    phi_array = phi_0 + np.linspace(-np.pi, np.pi, 1000)
# TODO: Make urf a function to avoid code duplication
    urf = (tomomachine.vrf1 * (np.cos(phi_0) - np.cos(phi_array))
           + (phi_0 - phi_array) * tomomachine.vrf1 * np.sin(phi_0))
    sync_time = tomomachine.synch_part_x * (time_array[1] - time_array[0])
    phi_array_matched = (time_array-sync_time) * (h_num*omega_rev0) + phi_0

    amplitude_array = np.linspace(0, np.min(urf), n_points_amplitude)
    summed_density_final = np.zeros(len(amplitude_array))
    local_density_final = np.zeros(len(amplitude_array))
    local_density_min_final = np.zeros(len(amplitude_array))
    local_density_max_final = np.zeros(len(amplitude_array))
    emittance_density = np.zeros(len(amplitude_array))

    for idx_amplitude, amplitude in enumerate(amplitude_array):

        phi_left = _urf_length_at_level(phi_array, urf, phi_0, amplitude)[-1]

# TODO: Make dE_height_array function to avoid code duplication
        dE_height_array = np.abs(np.sqrt(tomomachine.q * (gamma0 - 1/gamma0)
                                         * tomomachine.e_rest
                                         / (np.pi*tomomachine.h_num*eta0)
                                         * (tomomachine.vrf1
                                            * (np.cos(phi_array_matched)
                                               -np.cos(phi_left))
                                            + (phi_array_matched - phi_left)
                                            * tomomachine.vrf1*np.sin(phi_0))
                                         )
                                 )

        dE_height_array[np.isnan(dE_height_array)] = 0

        dE_resolution = dE_array[1] - dE_array[0]
        summed_density = 0
        n_points_local = 0
        local_density = 0
        local_density_min = np.inf
        local_density_max = 0

        for idx_time in range(len(time_array)):
            if dE_height_array[idx_time] == 0:
                continue

            good_idx = np.where((dE_array > -dE_height_array[idx_time])
                                *(dE_array < dE_height_array[idx_time]))[0]

            summed_density += np.trapz(tomo_image[idx_time, good_idx])

            cross_dE_top = np.interp(dE_height_array[idx_time],
                                     dE_array[good_idx[-1] : good_idx[-1]+2],
                                     tomo_image[idx_time,
                                                good_idx[-1] : good_idx[-1]+2])

            summed_density += np.trapz([tomo_image[idx_time, good_idx[-1]],
                                        cross_dE_top],
                                       dx=((dE_height_array[idx_time]
                                            -dE_array[good_idx[-1]])
                                           /dE_resolution))

            cross_dE_bottom = np.interp(-dE_height_array[idx_time],
                                        dE_array[good_idx[0]-1 : good_idx[0]+1],
                                        tomo_image[idx_time, good_idx[0]-1
                                                             : good_idx[0]+1])

            summed_density += np.trapz([cross_dE_bottom,
                                        tomo_image[idx_time, good_idx[0]]],
                                       dx=((dE_array[good_idx[0]]
                                            -(-dE_height_array[idx_time]))
                                            /dE_resolution))

            local_density += (cross_dE_top + cross_dE_bottom) / 2
            local_density_min = np.min([local_density_min, cross_dE_top,
                                        cross_dE_bottom])
            local_density_max = np.max([local_density_max, cross_dE_top,
                                        cross_dE_bottom])

            n_points_local += 1

        if n_points_local != 0:
            summed_density_final[idx_amplitude] = (summed_density
                                                   / np.sum(tomo_image))
            local_density_final[idx_amplitude] = local_density / n_points_local
            local_density_min_final[idx_amplitude] = local_density_min
            local_density_max_final[idx_amplitude] = local_density_max
            emittance_density[idx_amplitude] = (2 / (tomomachine.h_num
                                                     *omega_rev0)
                                                *np.trapz(dE_height_array,
                                                      dx=phi_array_matched[1]
                                                        -phi_array_matched[0]))

    if local_density_final[0] == 0:
        slope_extrapolate = ((local_density_final[2]-local_density_final[1])
                             / (emittance_density[2] - emittance_density[1]))
        origin_extrapolate = (local_density_final[1]
                              - slope_extrapolate
                              * emittance_density[1])
        local_density_final[0] = origin_extrapolate
        local_density_min_final[0] = local_density_final[0]
        local_density_max_final[0] = local_density_final[0]

    return (emittance_density, local_density_final, local_density_min_final,
            local_density_max_final, summed_density_final)


def tomo_weight_clipping(tomomachine: MachineABC, time_array: Iterable[float],
                         dE_array: Iterable[float], map_tomo_x: Iterable[float],
                         map_tomo_y: Iterable[float],
                         weight_tomo: Iterable[float], emittance_target: float,
                         n_points_amplitude: int=100, idx_frame: int=None)\
                                        -> Tuple[FloatArr, FloatArr, IntArray]:
    """


    Parameters
    ----------
    tomomachine : MachineABC
        The machine object used for the reconstruction.
    time_array : Iterable[float]
        The time axis of the phase space.
    dE_array : Iterable[float]
        The energy axis of the phase space.
    map_tomo_x : Iterable[float]
        TODO: DESCRIPTION.
    map_tomo_y : Iterable[float]
        TODO: DESCRIPTION.
    weight_tomo : Iterable[float]
        TODO: DESCRIPTION.
    emittance_target : float
        TODO: DESCRIPTION.
    n_points_amplitude : int, optional
        TODO: DESCRIPTION. The default is 100.
    idx_frame : int, optional
        The profile number to compute at.
        The default is None.
        If None, the machine reference frame is used.

    Returns
    -------
    Tuple[FloatArr, FloatArr, IntArray]
        TODO: DESCRIPTION.
    """

    if idx_frame is None:
        idx_frame = tomomachine.machine_ref_frame

    turn = int(idx_frame * tomomachine.dturns)

    new_weight_tomo = np.array(weight_tomo)

    phi_0 = tomomachine.phi0[turn]
    energy = tomomachine.e0[turn]
    eta0 = tomomachine.eta0[turn]
    h_num = tomomachine.h_num
    omega_rev0 = tomomachine.omega_rev0[turn]
    gamma0 = energy / tomomachine.e_rest

    phi_array = phi_0 + np.linspace(-np.pi, np.pi, 1000)
    urf = (tomomachine.vrf1 * (np.cos(phi_0)-np.cos(phi_array))
           + (phi_0-phi_array) * tomomachine.vrf1 * np.sin(phi_0))
    sync_time = tomomachine.synch_part_x * (time_array[1] - time_array[0])
    phi_array_matched = (time_array-sync_time) * (h_num*omega_rev0) + phi_0

    amplitude_array = np.linspace(0, np.min(urf), n_points_amplitude)

    dt_res = time_array[1] - time_array[0]
    dE_res = dE_array[1] - dE_array[0]

    for amplitude in amplitude_array:

        phi_left = _urf_length_at_level(phi_array, urf, phi_0, amplitude)[-1]

        dE_height_array = np.abs(np.sqrt(tomomachine.q * (gamma0 - 1/gamma0)
                                         * tomomachine.e_rest
                                         / (np.pi*tomomachine.h_num*eta0)
                                         * (tomomachine.vrf1
                                            * (np.cos(phi_array_matched)
                                               - np.cos(phi_left))
                                            + (phi_array_matched-phi_left)
                                            * tomomachine.vrf1 * np.sin(phi_0))
                                         )
                                 )

        dE_height_array[np.isnan(dE_height_array)] = 0

        emittance_density = (2 / (tomomachine.h_num*omega_rev0)
                             * np.trapz(dE_height_array,
                                        dx=phi_array_matched[1]
                                           - phi_array_matched[0]))

        if emittance_density >= emittance_target:
            break

    idx_minmax = np.array([], dtype=int)

    for idx_time in range(len(time_array) - 1):

        idx_out = np.where(
            ((map_tomo_x[:, idx_frame] >= ((time_array[idx_time]
                                            - time_array[0]) / dt_res))
             *(map_tomo_x[:, idx_frame] <= ((time_array[idx_time+1]
                                             - time_array[0]) / dt_res)))
            *((map_tomo_y[:, idx_frame] >= ((dE_height_array[idx_time]
                                             - dE_array[0]) / dE_res))
              | (map_tomo_y[:, idx_frame] <= ((-dE_height_array[idx_time]
                                               - dE_array[0]) / dE_res))))
        new_weight_tomo[idx_out] = 0

        idx_in = np.where(
            ((map_tomo_x[:, idx_frame] >= ((time_array[idx_time]
                                            - time_array[0]) / dt_res))
             * (map_tomo_x[:, idx_frame] <= ((time_array[idx_time + 1]
                                              - time_array[0]) / dt_res)))
            * ((map_tomo_y[:, idx_frame] < ((dE_height_array[idx_time]
                                             - dE_array[0]) / dE_res))
               * (map_tomo_y[:, idx_frame] > ((-dE_height_array[idx_time]
                                               - dE_array[0]) / dE_res))))[0]

        if len(idx_in) > 1:
            idx_max = idx_in[map_tomo_y[idx_in, idx_frame]
                             == np.max(map_tomo_y[idx_in, idx_frame])][0]
            idx_min = idx_in[map_tomo_y[idx_in, idx_frame]
                             == np.min(map_tomo_y[idx_in, idx_frame])][0]

            idx_minmax = np.append(idx_minmax, idx_min)
            idx_minmax = np.append(idx_minmax, idx_max)

        elif len(idx_in) == 1:
            idx_minmax.append(np.where(map_tomo_y
                                       == np.max(map_tomo_y[idx_in,
                                                            idx_frame]))[0][0])

    new_weight_tomo = (new_weight_tomo
                       / np.sum(new_weight_tomo)
                       * np.sum(weight_tomo))

    return new_weight_tomo, dE_height_array, idx_minmax



"""Module containing operations for modifying the data derived from the cpp code.

:Author(s): **Bernardo Abreu Figueiredo**
"""

from ..utils import tomo_config as conf

def make_phase_space(xp: conf.ndarray[conf.int32],
                     yp: conf.ndarray[conf.int32],
                     weights: conf.ndarray,
                     n_bins: int, result_cpu: bool = True) -> conf.ndarray:
    index = yp + xp * n_bins

    phase_space = conf.zeros(n_bins**2)
    conf.add.at(phase_space, index, weights)

    phase_space = phase_space.reshape((n_bins, n_bins))
    if result_cpu:
        return phase_space.get()
    else:
        return phase_space


"""Module containing the kick-and-drift algorithm with CuPy.

:Author(s): **Bernardo Abreu Figueiredo**
"""

import numpy as np
import cupy as cp
import logging
from typing import Tuple

log = logging.getLogger(__name__)

def drift_down(dphi: cp.ndarray,
               denergy: cp.ndarray, drift_coef: float,
               n_particles: int) -> cp.ndarray:
    dphi += drift_coef * denergy
    return dphi

def drift_up(dphi: cp.ndarray,
             denergy: cp.ndarray, drift_coef: float,
             n_particles: int) -> cp.ndarray:
    dphi -= drift_coef * denergy
    return dphi

def kick_down(dphi: cp.ndarray,
              denergy: cp.ndarray, rfv1: float, rfv2: float,
              phi0: float, phi12: float, h_ratio: float, n_particles: int,
              acc_kick: float) -> cp.ndarray:
    denergy -= rfv1 * cp.sin(dphi + phi0) \
                      + rfv2 * cp.sin(h_ratio * (dphi + phi0 - phi12)) - acc_kick
    return denergy

def kick_up(dphi: cp.ndarray,
            denergy: cp.ndarray, rfv1: float, rfv2: float,
            phi0: float, phi12: float, h_ratio: float, n_particles: int,
            acc_kick: float) -> cp.ndarray:
    denergy += rfv1 * cp.sin(dphi + phi0) \
                  + rfv2 * cp.sin(h_ratio * (dphi + phi0 - phi12)) - acc_kick
    return denergy

def kick_drift_up_simultaneously(dphi: cp.ndarray, denergy: cp.ndarray, drift_coef: float, rfv1: float, rfv2: float,
            phi0: float, phi12: float, h_ratio: float, n_particles: int, acc_kick: float) -> Tuple[cp.ndarray, cp.ndarray]:
    dphi -= drift_coef * denergy
    denergy += (rfv1 * cp.sin(dphi + phi0) \
                  + rfv2 * cp.sin(h_ratio * (dphi + phi0 - phi12)) - acc_kick)
    return dphi, denergy

def kick_drift_down_simultaneously(dphi: cp.ndarray, denergy: cp.ndarray, drift_coef: float, rfv1: float, rfv2: float,
            phi0: float, phi12: float, h_ratio: float, n_particles: int, acc_kick: float) -> Tuple[cp.ndarray, cp.ndarray]:
    denergy -= (rfv1 * cp.sin(dphi + phi0) \
                  + rfv2 * cp.sin(h_ratio * (dphi + phi0 - phi12)) - acc_kick)
    dphi += drift_coef * denergy
    return dphi, denergy

def kick_and_drift_cupy(xp: cp.ndarray, yp: cp.ndarray,
                   denergy: cp.ndarray, dphi: cp.ndarray,
                   rfv1: np.ndarray, rfv2: np.ndarray,
                   phi0: np.ndarray,
                   deltaE0: np.ndarray,
                   drift_coef: np.ndarray,
                   phi12: float,
                   h_ratio: float,
                   dturns: int,
                   rec_prof: int,
                   deltaturn: int,
                   nturns: int,
                   nparts: int,
                   fortran_flag,
                   callback) -> Tuple[cp.ndarray, cp.ndarray]:
    
    drift_coef = cp.asarray(drift_coef)
    phi0 = cp.asarray(phi0)
    deltaE0 = cp.asarray(deltaE0)
    rfv1 = cp.asarray(rfv1)
    rfv2 = cp.asarray(rfv2)

    phi12_arr = cp.full(nturns+1, phi12)
    # Preparation end

    profile = rec_prof
    turn = rec_prof * dturns + deltaturn

    if deltaturn < 0:
        profile -= 1

    # Value-based copy to avoid side-effects
    xp[profile] = cp.copy(dphi)
    yp[profile] = cp.copy(denergy)

    while turn < nturns:
        turn += 1
        dphi, denergy = kick_drift_up_simultaneously(dphi, denergy, drift_coef[turn-1], rfv1[turn], rfv2[turn],
                                                        phi0[turn], phi12_arr[turn], h_ratio, nparts, deltaE0[turn])

        if turn % dturns == 0:
            profile += 1

            xp[profile] = cp.copy(dphi)
            yp[profile] = cp.copy(denergy)

    profile = rec_prof
    turn = rec_prof * dturns

    if profile > 0:
        # going back to initial coordinates
        dphi = cp.copy(xp[rec_prof])
        denergy = cp.copy(yp[rec_prof])

        # Downwards
        while turn > 0:
            dphi, denergy = kick_drift_down_simultaneously(dphi, denergy, drift_coef[turn-1], rfv1[turn], rfv2[turn],
                                                                    phi0[turn], phi12_arr[turn], h_ratio, nparts, deltaE0[turn])
            turn -= 1

            if (turn % dturns == 0):
                profile -= 1
                xp[profile] = cp.copy(dphi)
                yp[profile] = cp.copy(denergy)


# General imports
import os

import matplotlib.pyplot as plt
import numpy as np

# Tomo imports
import longitudinal_tomography.utils.tomo_run as tomorun

ex_dir = os.path.split(os.path.realpath(os.path.dirname(__file__)))[0]
in_file_pth = os.path.join(ex_dir, 'input_files', 'flatTopINDIVRotate2.dat')

tRange, ERange, density = tomorun.run(in_file_pth)

# %%

vmin = np.min(density[density > 0])
vmax = np.max(density)

plt.contourf(tRange * 1E9, ERange / 1E6, density.T,
             levels=np.linspace(vmin, vmax, 50), cmap='Oranges')
plt.xlabel('dt (ns)')
plt.ylabel('dE (MeV)')
plt.show()



"""
Compatibility module. Everything in this submodule can be removed at once with
minimal impact to the rest of the package (some if-statements need to be
removed in the rest of the package as well).

The intention of this package is to provide Fortran I/O compatibility with the
tomoscope until it is deprecated.

:Author(s): **Anton Lu**
"""

import logging
import os
from typing import TYPE_CHECKING

import numpy as np

from .. import assertions as asrt, exceptions as expt

if TYPE_CHECKING:
    from ..tracking.machine import Machine
    from ..tracking.particles import Particles

log = logging.getLogger(__name__)


# Input-output
# --------------------------------------------------------------- #
#                           PROFILES                              #
# --------------------------------------------------------------- #

def save_profile(profiles: np.ndarray, recprof: int, output_dir: str):
    """Write phase-space image to text-file in the original format.
    The name of the file will be profileXXX.data, where XXX is the index
    of the time frame to be reconstructed counting from one.

    Parameters
    ----------
    profiles: ndarray
        Profile measurements as a 2D array with the shape: (nprofiles, nbins).
    recprof: int
        Index of profile to be saved.
    output_dir: string
        Path to output directory.
    """
    out_profile = profiles[recprof].flatten()
    file_path = os.path.join(output_dir, f'profile{recprof + 1:03d}.data')
    with open(file_path, 'w') as f:
        for element in out_profile:
            f.write(f' {element:0.7E}\n')


def save_self_volt_profile(self_fields: np.ndarray, output_dir: str):
    """Write self volts to text file in tomoscope format.

    Parameters
    ----------
    self_fields: ndarray
        Calculated self-field voltages.
    output_dir: string
        Path to output directory.
    """
    out_profile = self_fields.flatten()
    file_path = os.path.join(output_dir, 'vself.data')
    with open(file_path, 'w') as f:
        for element in out_profile:
            f.write(f' {element:0.7E}\n')


# --------------------------------------------------------------- #
#                         PHASE-SPACE                             #
# --------------------------------------------------------------- #

def save_phase_space(image: np.ndarray, recprof: int, output_path: str):
    """Save phase-space image in a tomoscope format.

    Parameters
    ----------
    image: ndarray
        2D array holding the weight of each cell of the recreated phase-space.
    recprof: int
        Index of reconstructed profile.
    output_path: string
        Path to the output directory.
    """
    log.info(f'Saving image{recprof} to {output_path}')
    image = image.flatten()
    file_path = os.path.join(output_path, f'image{recprof + 1:03d}.data')
    with open(file_path, 'w') as f:
        for element in image:
            f.write(f'  {element:0.7E}\n')


# --------------------------------------------------------------- #
#                           PROFILES                              #
# --------------------------------------------------------------- #


# --------------------------------------------------------------- #
#                          PLOT INFO                              #
# --------------------------------------------------------------- #

def write_plotinfo(machine: 'Machine', particles: 'Particles',
                   profile_charge: float) -> str:
    """Creates string of plot info needed for the original output
    for the tomography program.

    Parameters
    ----------
    machine: Machine
        Object containing machine parameters.
    particles: Particles
        Object containing particle distribution and information about
        the phase space reconstruction.
    profile_charge: float
        Total charge of a reference profile.

    Returns
    -------
    plot_info: string
        String containing information needed by the tomoscope application.
        The returned string has the same format as in the original version.

    """
    recprof = machine.filmstart
    rec_turn = recprof * machine.dturns

    # Check if a Fortran styled fit has been performed.
    fit_performed = True
    fit_info_vars = ['fitted_synch_part_x', 'bunchlimit_low', 'bunchlimit_up']
    for var in fit_info_vars:
        if not hasattr(machine, var) or getattr(machine, var) is None:
            fit_performed = False
            break

    if fit_performed:
        bunchlimit_low = machine.bunchlimit_low
        bunchlimit_up = machine.bunchlimit_up
        fitted_synch_part_x = machine.fitted_synch_part_x
    else:
        bunchlimit_low = 0.0
        bunchlimit_up = 0.0
        fitted_synch_part_x = 0.0

    if particles.dEbin is None:
        raise expt.EnergyBinningError(
            'dEbin has not been calculated for this '
            'phase space info object.\n'
            'Cannot print plot info.')
    if particles.imin is None or particles.imax is None:
        raise expt.PhaseLimitsError(
            'The limits in phase (I) has not been found '
            'for this phase space info object.\n'
            'Cannot print plot info.')

        # '+ 1': Converting from Python to Fortran indexing
    out_s = f' plotinfo.data\n' \
            f'Number of profiles used in each reconstruction,\n' \
            f' profilecount = {machine.nprofiles}\n' \
            f'Width (in pixels) of each image = ' \
            f'length (in bins) of each profile,\n' \
            f' profilelength = {machine.nbins}\n' \
            f'Width (in s) of each pixel = width of each profile bin,\n' \
            f' dtbin = {machine.dtbin:0.4E}\n' \
            f'Height (in eV) of each pixel,\n' \
            f' dEbin = {particles.dEbin:0.4E}\n' \
            f'Number of elementary charges in each image,\n' \
            f' eperimage = ' \
            f'{profile_charge:0.3E}\n' \
            f'Position (in pixels) of the reference synchronous point:\n' \
            f' xat0 =  {machine.synch_part_x:.3f}\n' \
            f' yat0 =  {machine.synch_part_y:.3f}\n' \
            f'Foot tangent fit results (in bins):\n' \
            f' tangentfootl =    {bunchlimit_low:.3f}\n' \
            f' tangentfootu =    {bunchlimit_up:.3f}\n' \
            f' fit xat0 =   {fitted_synch_part_x:.3f}\n' \
            f'Synchronous phase (in radians):\n' \
            f' phi0( {recprof + 1}) = {machine.phi0[rec_turn]:.4f}\n' \
            f'Horizontal range (in pixels) of the region in ' \
            f'phase space of map elements:\n' \
            f' imin( {recprof + 1}) =   {particles.imin} and ' \
            f'imax( {recprof + 1}) =  {particles.imax}'
    return out_s


# --------------------------------------------------------------- #
#                         DISCREPANCY                             #
# --------------------------------------------------------------- #

def save_difference(diff: np.ndarray, output_path: str, recprof: int):
    """Write reconstruction discrepancy to text file with original format.

    Parameters
    ----------
    diff: ndarray
        1D array containing the discrepancy for the phase space at each
        iteration of the reconstruction.
    output_path: string
        Path to output directory.
    recprof: int
        Index of profile to be saved.
    """
    log.info(f'Saving saving difference to {output_path}')
    file_path = os.path.join(output_path, f'd{recprof + 1:03d}.data')
    with open(file_path, 'w') as f:
        for i, d in enumerate(diff):
            f.write(f'         {i:3d}  {d:0.7E}\n')


# --------------------------------------------------------------- #
#                          TRACKING                               #
# --------------------------------------------------------------- #

def print_tracking_status(ref_prof: int, to_profile: int):
    """Write output for particle tracking in the original format.
    Since the original algorithm is somewhat different,
    the **output concerning lost particles is not valid**.
    Meanwhile, the format it is needed by the tomoscope application.
    Profile numbers are added by one in order to compensate for
    differences in Python and Fortran indexing. Fortran counts from
    one, Python counts from 0.
    This function is used in the tracking algorithm.

    Parameters
    ----------
    ref_prof: int
        Index of profile to be reconstructed.
    to_profile: int
        Profile to which the algorithm is currently tracking towards.
    """
    print(f' Tracking from time slice  {ref_prof + 1} to  '
          f'{to_profile + 1},   0.000% went outside the image width.')


# Data treatment

def calc_baseline(waterfall: np.ndarray, ref_prof: int,
                  percent: float = 0.05) -> float:
    """Function for finding baseline of raw data.

    The function is based on the original Fortran program,
    and uses a percentage of a reference profile in order to
    find the baseline of the measurements.

    Parameters
    ----------
    waterfall: ndarray
        Raw-data shaped as waterfall: (nprofiles, nbins).
    ref_prof: int
        Index of reference profile.
    percent: float, optional, default=0.05
        A number between 0 and 1 describing the percentage of the
        reference profile used to find the baseline.

    Returns
    -------
    baseline: float
        Baseline of reference profile.

    Raises
    ------
    InputError: Exception
        Raised if percentage is not given as a float between 0 and 1.

    """
    asrt.assert_inrange(percent, 'percent', 0.0, 1.0, expt.InputError,
                        'The chosen percent of raw_data '
                        'to create baseline from is not valid')

    nbins = len(waterfall[ref_prof])
    iend = int(percent * nbins)

    return np.sum(waterfall[ref_prof, :iend]) / np.floor(percent * nbins)



"""
Tomoscope specific codes

:Author(s): **Anton Lu**
"""
import logging
from os import path

import numpy as np

from ..utils import tomo_output as tomoout


log = logging.getLogger(__name__)


def save_difference(diff: np.ndarray, output_path: str, film: int):
    # Saving to file with numbers counting from one
    log.info(f'Saving saving difference to {output_path}')

    with open(path.join(output_path, f'd{film + 1:03d}.data'), 'w') as f:
        for i, d in enumerate(diff):
            if i < 10:
                f.write(f'           {i}  {d:0.7E}\n')
            else:
                f.write(f'          {i}  {d:0.7E}\n')


def save_image(xp: np.ndarray, yp: np.ndarray, weight: np.ndarray,
               n_bins: int, film: int, output_path: str):
    # Creating n_bins * n_bins phase-space image
    log.info(f'Saving picture {film}.')

    phase_space = tomoout.create_phase_space_image(xp, yp, weight, n_bins,
                                                   film)

    log.info(f'Saving image{film} to {output_path}')
    out_ps = phase_space.flatten()
    with open(path.join(output_path, f'image{film + 1:03d}.data'), 'w') as f:
        for element in out_ps:
            f.write(f'  {element:0.7E}\n')


def save_profile(prof: np.ndarray, film: int, output_path: str):

    log.info(f'Saving profile{film} to {output_path}')

    with open(path.join(output_path, f'profile{film + 1:03d}.data'), 'w') as f:
        for element in prof:
            f.write(f' {element:0.7E}\n')



"""Unit-tests for the Profiles class.

Run as python test_tracking.py in console or via coverage
"""
from __future__ import annotations
import os
import unittest

import numpy as np
import numpy.testing as nptest
import scipy.signal as sig

from .. import commons
import longitudinal_tomography.data.profiles as prf
import longitudinal_tomography.tracking.machine as mch
from longitudinal_tomography import exceptions as expt

# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestProfiles(unittest.TestCase):

    def test_create_waterfall_no_iter_fails(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()
        waterfall = 1

        with self.assertRaises(expt.WaterfallError,
                               msg='Waterfall as non-iterable should '
                                   'raise an exception'):
            profiles = prf.Profiles(machine, machine.dtbin, waterfall)

    def test_create_waterfall_wrong_shape_fails(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()
        waterfall = np.ones((30, 100))

        with self.assertRaises(expt.WaterfallError,
                               msg='Waterfall as having the wrong amount '
                                   'of profiles should raise an exception'):
            profiles = prf.Profiles(machine, machine.dtbin, waterfall)

    def test_create_waterfall_reduced_to_zeros_fail(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()
        waterfall = np.ones((150, 100)) * -1

        with self.assertRaises(expt.WaterfallReducedToZero,
                               msg='Waterfall as having the wrong amount '
                                   'of profiles should raise an exception'):
            profiles = prf.Profiles(machine, machine.dtbin, waterfall)

    def test_calc_profile_charge_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()
        waterfall = self._load_waterfall()

        profiles = prf.Profiles(machine, machine.dtbin, waterfall)
        profiles.calc_profilecharge()

        # Move decimal to improve comparal
        correct_prof_charge = 206096981027.60077 * 1.0e-11
        profiles.profile_charge *= 1.0e-11

        self.assertAlmostEqual(
            profiles.profile_charge, correct_prof_charge,
            msg='The profile charge was calculated incorrectly')

    def test_calc_self_fields_no_prof_charge_fails(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()
        waterfall = self._load_waterfall()

        profiles = prf.Profiles(machine, machine.dtbin, waterfall)
        with self.assertRaises(expt.ProfileChargeNotCalculated,
                               msg='An exception should be raised when '
                                   'an attempt is made to calculate the '
                                   'self-fields of a bunch, whitout first '
                                   'having provided the profile charge.'):
            profiles.calc_self_fields()

    def test_calc_self_fields_original_correct_bdot_greater_than_zero(self):
        waterfall = self._load_waterfall()
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        sample_time = machine.dtbin

        # Update fields due to loading of rebinned waterfall.
        rbn = 3
        machine.dtbin *= rbn
        machine.synch_part_x /= rbn

        # Set space charge parameters
        machine.g_coupling = 1.0
        machine.zwall_over_n = 50.0

        profiles = prf.Profiles(machine, sample_time, waterfall)
        profiles.calc_profilecharge()
        profiles.calc_self_fields()

        correct_vself = self._load_vself()
        correct_phiwrap = 6.283185307179586
        correct_wrap_length = 455

        self.assertAlmostEqual(profiles.phiwrap, correct_phiwrap,
                               msg='phiwrap calculated incorrectly.')
        self.assertEqual(profiles.wrap_length, correct_wrap_length,
                         msg='wrap length calculated incorrectly.')
        nptest.assert_almost_equal(
            profiles.vself, correct_vself,
            err_msg='Error in calculation self fields using original '
                    'method and filter.')

    def test_calc_self_fields_wraplength_correct_bdot_less_than_zero(self):
        waterfall = self._load_waterfall()
        machine = mch.Machine(**MACHINE_ARGS)
        machine.bdot = 0.0
        machine.values_at_turns()

        sample_time = machine.dtbin

        # Update fields due to loading of rebinned waterfall.
        rbn = 3
        machine.dtbin *= rbn
        machine.synch_part_x /= rbn

        # Set space charge parameters
        machine.g_coupling = 1.0
        machine.zwall_over_n = 50.0

        profiles = prf.Profiles(machine, sample_time, waterfall)
        profiles.calc_profilecharge()
        profiles.calc_self_fields()

        correct_vself = self._load_vself()
        correct_phiwrap = 6.283185307179586
        correct_wrap_length = 457

        self.assertAlmostEqual(profiles.phiwrap, correct_phiwrap,
                               msg='phiwrap calculated incorrectly.')
        self.assertEqual(profiles.wrap_length, correct_wrap_length,
                         msg='wrap length calculated incorrectly.')

    def test_calc_self_fields_filter_outside_correct(self):
        waterfall = self._load_waterfall()
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        sample_time = machine.dtbin

        # Update fields due to loading of rebinned waterfall.
        rbn = 3
        machine.dtbin *= rbn
        machine.synch_part_x /= rbn

        # Set space charge parameters
        machine.g_coupling = 1.0
        machine.zwall_over_n = 50.0

        profiles = prf.Profiles(machine, sample_time, waterfall)
        profiles.calc_profilecharge()

        smoothed_profs = np.copy(waterfall)
        smoothed_profs /= np.sum(smoothed_profs, axis=1)[:, None]
        smoothed_profs = sig.savgol_filter(
                                x=smoothed_profs, window_length=7,
                                polyorder=4, deriv=1)

        profiles.calc_self_fields(filtered_profiles=smoothed_profs)

        correct_vself = self._load_vself()
        correct_phiwrap = 6.283185307179586
        correct_wrap_length = 455

        self.assertAlmostEqual(profiles.phiwrap, correct_phiwrap,
                               msg='phiwrap calculated incorrectly.')
        self.assertAlmostEqual(profiles.wrap_length, correct_wrap_length,
                               msg='wrap length calculated incorrectly.')
        nptest.assert_almost_equal(
            profiles.vself, correct_vself,
            err_msg='Error in calculation self fields using original '
                    'method, filtered profiles provided via parameters')

    def test_calc_self_fields_filter_outside_no_iter_fails(self):
        waterfall = self._load_waterfall()
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        sample_time = machine.dtbin

        profiles = prf.Profiles(machine, sample_time, waterfall)
        profiles.calc_profilecharge()

        with self.assertRaises(
                expt.FilteredProfilesError,
                msg='Providing the filtered profiles as non-iterable object '
                    'should raise an exception'):
            profiles.calc_self_fields(filtered_profiles=1)

    def test_calc_self_fields_filter_outside_bad_shape_fails(self):
        waterfall = self._load_waterfall()
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        sample_time = machine.dtbin

        profiles = prf.Profiles(machine, sample_time, waterfall)
        profiles.calc_profilecharge()

        with self.assertRaises(
                expt.FilteredProfilesError,
                msg='Providing the filtered profiles not having the '
                    'same shape as the waterfall should raise an exception'):
            profiles.calc_self_fields(filtered_profiles=np.ones((60, 40)))

    def _load_waterfall(self):
        base_dir = os.path.split(os.path.realpath(__file__))[0]
        base_dir = os.path.split(base_dir)[0]
        data_path = os.path.join(base_dir, 'resources')

        waterfall = np.load(os.path.join(
            data_path, 'waterfall_INDIVShavingC325.npy'))
        return waterfall

    def _load_vself(self):
        base_dir = os.path.split(os.path.realpath(__file__))[0]
        base_dir = os.path.split(base_dir)[0]
        data_path = os.path.join(base_dir, 'resources')
        vself = np.load(os.path.join(
                        data_path, 'vself_INDIVShavingC325.npy'))
        return vself



"""Module containing Python wrappers for C++ functions.

Should only be used by advanced users.

:Author(s): **Christoffer Hjertø Grindheim**
"""

import ctypes as ct
import logging
import os
import sys
from glob import glob
from typing import Tuple, TYPE_CHECKING
import tomo.cpp_routines.libtomo as libtomo

import numpy as np

from .. import exceptions as expt

if TYPE_CHECKING:
    from ..tracking.machine import Machine

log = logging.getLogger(__name__)


# =============================================================
# Functions for particle tracking
# =============================================================


def kick(machine: 'Machine', denergy: np.ndarray, dphi: np.ndarray,
         rfv1: np.ndarray, rfv2: np.ndarray, npart: int, turn: int,
         up: bool = True) -> np.ndarray:
    """Wrapper for C++ kick function.

    Particle kick for **one** machine turn.

    Used in the :mod:`longitudinal_tomography.tracking.tracking` module.

    Parameters
    ----------
    machine: Machine
        Object holding machine parameters.
    denergy: ndarray
        1D array holding the energy difference relative to the synchronous
        particle for each particle at a turn.
    dphi: ndarray
        1D array holding the phase difference relative to the synchronous
        particle for each particle at a turn.
    rfv1: ndarray
        1D array holding the radio frequency voltage at RF station 1 for
        each turn, multiplied with the charge state of the particles.
    rfv2: ndarray
        1D array holding the radio frequency voltage at RF station 2 for
        each turn, multiplied with the charge state of the particles.
    npart: int
        The number of tracked particles.
    turn: int
        The current machine turn.
    up: boolean, optional, default=True
        Direction of tracking. Up=True tracks towards last machine turn,
        up=False tracks toward first machine turn.

    Returns
    -------
    denergy: ndarray
        1D array containing the new energy of each particle after voltage kick.
    """
    log.warning('tomo.cpp_routines.tomolib_wrappers.kick has moved to '
                'tomo.cpp_routines.libtomo.kick with the same '
                'function and return signature. This function is '
                'provided for backwards compatibility and can be '
                'removed without further notice.')
    return libtomo.kick(machine, denergy, dphi, rfv1, rfv2,
                        npart, turn, up)


def drift(denergy: np.ndarray, dphi: np.ndarray, drift_coef: np.ndarray,
          npart: int, turn: int, up: bool = True) -> np.ndarray:
    """Wrapper for C++ drift function.

    Particle drift for **one** machine turn

    Used in the :mod:`~longitudinal_tomography.tracking.tracking` module.

    Parameters
    ----------
    denergy: ndarray
        1D array holding the energy difference relative to the synchronous
        particle for each particle at a turn.
    dphi: ndarray
        1D array holding the phase difference relative to the synchronous
        particle for each particle at a turn.
    drift_coef: ndarray
        1D array of drift coefficient at each machine turn.
    npart: int
        The number of tracked particles.
    turn: int
        The current machine turn.
    up: boolean, optional, default=True
        Direction of tracking. Up=True tracks towards last machine turn,
        up=False tracks toward first machine turn.

    Returns
    -------
    dphi: ndarray
        1D array containing the new phase for each particle
        after drifting for a machine turn.
    """
    log.warning('tomo.cpp_routines.tomolib_wrappers.drift has moved to '
                'tomo.cpp_routines.libtomo.drift with the same '
                'function and return signature. This function is '
                'provided for backwards compatibility and can be '
                'removed without further notice.')
    return libtomo.drift(denergy, dphi, drift_coef, npart, turn, up)


def kick_and_drift(xp: np.ndarray, yp: np.ndarray,
                   denergy: np.ndarray, dphi: np.ndarray,
                   rfv1: np.ndarray, rfv2: np.ndarray, rec_prof: int,
                   nturns: int, nparts: int,
                   phi0: np.ndarray = None,
                   deltaE0: np.ndarray = None,
                   omega_rev0: np.ndarray = None,
                   drift_coef: np.ndarray = None,
                   phi12: float = None,
                   h_ratio: float = None,
                   dturns: int = None,
                   machine: 'Machine' = None,
                   ftn_out: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    """Wrapper for full kick and drift algorithm written in C++.

    Tracks all particles from the time frame to be recreated,
    trough all machine turns.

    Used in the :mod:`longitudinal_tomography.tracking.tracking` module.

    Parameters
    ----------
    xp: ndarray
        2D array large enough to hold the phase of each particle
        at every time frame. Shape: (nprofiles, nparts)
    yp: ndarray
        2D array large enough to hold the energy of each particle
        at every time frame. Shape: (nprofiles, nparts)
    denergy: ndarray
        1D array holding the energy difference relative to the synchronous
        particle for each particle the initial turn.
    dphi: ndarray
        1D array holding the phase difference relative to the synchronous
        particle for each particle the initial turn.
    rfv1: ndarray
        Array holding the radio frequency voltage at RF station 1 for each
        turn, multiplied with the charge state of the particles.
    rfv2: ndarray
        Array holding the radio frequency voltage at RF station 2 for each
        turn, multiplied with the charge state of the particles.
    rec_prof: int
        Index of profile to be reconstructed.
    nturns: int
        Total number of machine turns.
    nparts: int
        The number of particles.
    args: tuple
        Arguments can be provided via the args if a machine object is not to
        be used. In this case, the args should be:

        - phi0
        - deltaE0
        - omega_rev0
        - drift_coef
        - phi12
        - h_ratio
        - dturns

        The args will not be used if a Machine object is provided.

    machine: Machine, optional, default=False
        Object containing machine parameters.
ftn_out: boolean, optional, default=False
        Flag to enable printing of status of tracking to stdout.
        The format will be similar to the Fortran version.
        Note that the **information regarding lost particles
        are not valid**.

    Returns
    -------
    xp: ndarray
        2D array holding every particles coordinates in phase [rad]
        at every time frame. Shape: (nprofiles, nparts)
    yp: ndarray
        2D array holding every particles coordinates in energy [eV]
        at every time frame. Shape: (nprofiles, nparts)
    """

    log.warning('tomo.cpp_routines.tomolib_wrappers.kick_and_drift has moved '
                'to tomo.cpp_routines.libtomo.kick_and_drift with the same '
                'function and return signature. This function is '
                'provided for backwards compatibility and can be '
                'removed without further notice.')
    machine_args = [phi0, deltaE0, omega_rev0,
                    drift_coef, phi12, h_ratio, dturns]

    if machine is not None:
        return libtomo.kick_and_drift(xp, yp, denergy, dphi, rfv1, rfv2,
                                      machine, rec_prof, nturns, nparts,
                                      ftn_out)
    elif all([x is not None for x in machine_args]):
        return libtomo.kick_and_drift(xp, yp, denergy, dphi, rfv1, rfv2,
                                      phi0, deltaE0, drift_coef, phi12,
                                      h_ratio, dturns, rec_prof, nturns,
                                      nparts, ftn_out)
    else:
        raise expt.InputError(
            'Wrong input arguments.\n'
            'Either: phi0, deltaE0, omega_rev0, '
            'drift_coef, phi12, h_ratio, dturns '
            'OR machine is required.')


# =============================================================
# Functions for phase space reconstruction
# =============================================================


def back_project(weights: np.ndarray, flat_points: np.ndarray,
                 flat_profiles: np.ndarray, nparts: int, nprofs: int) \
        -> np.ndarray:
    """Wrapper for back projection routine written in C++.
    Used in the :mod:`~longitudinal_tomography.tomography.tomography` module.

    Parameters
    ----------
    weights: ndarray
        1D array containing the weight of each particle.
    flat_points: ndarray
        2D array containing particle coordinates as integers, pointing
        at flattened versions of the waterfall. Shape: (nparts, nprofiles)
    flat_profiles: ndarray
        1D array containing a flattened waterfall.
    nparts: int
        Number of tracked particles.
    nprofs: int
        Number of profiles.

    Returns
    -------
    weights: ndarray
        1D array containing the **new weight** of each particle.
    """
    log.warning('tomo.cpp_routines.tomolib_wrappers.back_project has moved to '
                'tomo.cpp_routines.libtomo.back_project with the same '
                'function and return signature. This function is '
                'provided for backwards compatibility and can be '
                'removed without further notice.')
    return libtomo.back_project(weights, flat_points, flat_profiles, nparts,
                                nprofs)


def project(recreated: np.ndarray, flat_points: np.ndarray,
            weights: np.ndarray,
            nparts: int, nprofs: int, nbins: int) -> np.ndarray:
    """Wrapper projection routine written in C++.
    Used in the :mod:`~longitudinal_tomography.tomography.tomography` module.

    Parameters
    ----------
    recreated: ndarray
        2D array with the shape of the waterfall to be recreated,
        initiated as zero. Shape: (nprofiles, nbins)
    flat_points: ndarray
        2D array containing particle coordinates as integers, pointing
        at flattened versions of the waterfall. Shape: (nparts, nprofiles)
    weights: ndarray
        1D array containing the weight of each particle.
    nparts: int
        Number of tracked particles.
    nprofs: int
        Number of profiles.
    nbins: int
        Number of bins in profiles.

    Returns
    -------
    recreated: ndarray
        2D array containing the projected profiles as waterfall.
        Shape: (nprofiles, nbins)
    """
    log.warning('tomo.cpp_routines.tomolib_wrappers.project has moved to '
                'tomo.cpp_routines.libtomo.project with the same '
                'function and return signature. This function is '
                'provided for backwards compatibility and can be '
                'removed without further notice.')
    return libtomo.project(recreated, flat_points, weights, nparts, nprofs,
                           nbins)


def reconstruct(xp: np.ndarray, waterfall: np.ndarray, niter: int, nbins: int,
                npart: int, nprof: int, verbose: bool):
    """Wrapper for full reconstruction in C++.
    Used in the :mod:`~longitudinal_tomography.tomography.tomography` module.

    Parameters
    ----------
    xp: ndarray
        2D array containing the coordinate of each particle
        at each time frame. Coordinates should given be as integers of
        the phase space coordinate system. shape: (nparts, nprofiles).
    waterfall: ndarray
        2D array containing measured profiles as waterfall.
        Shape: (nprofiles, nbins).
    niter: int
        Number of iterations in the reconstruction process.
    nbins: int
        Number of bins in a profile.
    npart: int
        Number of tracked particles.
    nprof: int
        Number of profiles.
    verbose: boolean
        Flag to indicate that the tomography routine should broadcast its
        status to stdout. The output is identical to the output
        from the Fortran version.

    Returns
    -------
    weights: ndarray
        1D array containing the weight of each particle.
    discr: ndarray
        1D array containing discrepancy at each
        iteration of the reconstruction.
    recreated: ndarray
        2D array containing the projected profiles as waterfall.
        Shape: (nprofiles, nbins)
    """
    log.warning('tomo.cpp_routines.tomolib_wrappers.reconstruct has moved to '
                'tomo.cpp_routines.libtomo.reconstruct with the same '
                'function and return signature. This function is '
                'provided for backwards compatibility and can be '
                'removed without further notice.')
    return libtomo.reconstruct(xp, waterfall, niter, nbins, npart, nprof,
                               verbose)



"""
Macro functions provided for convenience

:Author(s): **Anton Lu**
"""
from __future__ import annotations
import typing as t
import numpy as np

from .tracking import Tracking
from .tracking.machine_base import MachineABC
from .tracking import particles
from .tomography import tomography
from .utils import tomo_input as tin
from .data import data_treatment as dtreat


__all__ = ['track', 'tomogram']


def read_input_file(input_file_path: str) -> t.Tuple[MachineABC, np.ndarray]:
    """
    Read input file and return a machine and a waterfall

    :param input_file_path: path to input file
    :return: machine and waterfall
    """
    raw_params, raw_data = tin.get_user_input(input_file_path)

    machine, frames = tin.txt_input_to_machine(raw_params)
    machine.values_at_turns()
    measured_waterfall = frames.to_waterfall(raw_data)

    profiles = tin.raw_data_to_profiles(
        measured_waterfall, machine, frames.rebin, frames.sampling_time)
    profiles.calc_profilecharge()

    if profiles.machine.synch_part_x < 0:
        fit_info = dtreat.fit_synch_part_x(profiles)
        machine.load_fitted_synch_part_x_ftn(fit_info)

    return machine, profiles.waterfall


def track(machine: MachineABC, reconstruction_idx: int = None,
          callback: t.Callable = None) \
        -> t.Tuple[np.ndarray, np.ndarray]:

    tracker = Tracking(machine)
    dphi, denergy = tracker.track(reconstruction_idx, callback=callback)
    xp, yp = particles.physical_to_coords(dphi, denergy, machine,
                                          tracker.particles.xorigin,
                                          tracker.particles.dEbin)
    xp, yp = particles.ready_for_tomography(xp, yp, machine.nbins)

    return xp, yp


def tomogram(waterfall: np.ndarray, xp: np.ndarray, yp: np.ndarray,
             n_iter: int, callback: t.Callable = None) \
        -> tomography.Tomography:

    tomo = tomography.Tomography(waterfall, xp, yp)
    tomo.run(n_iter, callback=callback)

    return tomo



"""Module containing functions for handling input from
Fortran style text files.

:Author(s): **Christoffer Hjertø Grindheim**, **Anton Lu**
"""
from __future__ import annotations
import os
import sys
import typing as t

import numpy as np

from .. import assertions as asrt, exceptions as expt
from ..data import pre_process
from ..data.profiles import Profiles
from ..tracking.machine import Machine
from ..compat import fortran

# Some constants for input files containing machine parameters.
from ..tracking.machine_base import MachineABC

PARAMETER_LENGTH = 98
RAW_DATA_FILE_IDX = 12
OUTPUT_DIR_IDX = 14


class Frames:
    """Class for storing raw data and information on how to treat them,
    based on a Fortran style input file.

    Parameters
    ----------
    framecount: int
        Number of time frames.
    framelength: int
        Number of bins in a time frame.
    skip_frames: int
        Number of time frames to ignore from the
        beginning of the raw input file.
    skip_bins_start: int
        Subtract this number of bins from start of the raw input.
    skip_bins_end: int
        Subtract this number of bins from end of the raw input.
    rebin: int
        Rebinning factor - Number of frame bins to rebin into one profile bin.
    dtbin: float
        Size of profile bins (sampling size) [s].
    raw_data_path: string, optional, default=''
        Path to file holding raw data for programmers reference.

    Attributes
    ----------
    raw_data:
        Measured raw data as read from text file.
    nframes: int
        Number of measured time frames.
    nbins_frame: int
        Number of bins in a measured time frame.
    skip_frames: int
        Number of time frames to ignore from the
        beginning of the raw input file.
    skip_bins_start: int
        Subtract this number of bins from start of the raw input.
    skip_bins_end: int
        Subtract this number of bins from end of the raw input.
    rebin: int
        Rebinning factor, the number of frame bins to
        rebin into one profile bin.
    sampling_time: float
        Size of profile bins [s].
    raw_data_path: string
        Path to file holding raw data for programmers reference.
    """

    def __init__(self, framecount: int, framelength: int, skip_frames: int,
                 skip_bins_start: int, skip_bins_end: int, rebin: int,
                 dtbin: float, raw_data_path: str = ''):
        self.raw_data_path = raw_data_path
        self.nframes = framecount
        self.nbins_frame = framelength
        self.skip_frames = skip_frames
        self.skip_bins_start = skip_bins_start
        self.skip_bins_end = skip_bins_end
        self.rebin = rebin
        self.sampling_time = dtbin

        self._raw_data: np.ndarray = None

    @property
    def raw_data(self) -> t.Union[np.ndarray, None]:
        """Raw data defined as a @property.

        Holds assertions for validity of raw data.

        Parameters
        ----------
        in_raw_data: ndarray
            1D array holding raw data as a direct read from text file.

        Returns
        -------
        raw_data: ndarray, none
            Copy of array holding raw data as a direct read from text file.
            None is returned if object holds no raw data.

        Raises
        ------
        RawDataImportError: Exception
            Raised if raw data is not iterable or has an unexpected length.

        """
        if self._raw_data is not None:
            return np.copy(self._raw_data)
        else:
            return None

    @raw_data.setter
    def raw_data(self, in_raw_data: t.Collection):
        if not hasattr(in_raw_data, '__iter__'):
            raise expt.RawDataImportError('Raw data should be iterable')

        ndata = self.nframes * self.nbins_frame
        if len(in_raw_data) == ndata:
            self._raw_data = np.array(in_raw_data)
        else:
            raise expt.RawDataImportError(
                f'Raw data has length {len(in_raw_data)}.\n'
                f'expected length: {ndata}')

    def nprofs(self) -> int:
        """Function for calculating the number of profiles.
        This number should be used when creating a machine object.

        Returns
        -------
        nprofiles: int
            Number of profiles.
        """
        return self.nframes - self.skip_frames

    def nbins(self) -> int:
        """Function to calculate number of bins in profiles.
        This number should be used when creating a machine object.

        Returns
        -------
        nbins: int
            Number of bins in profiles.
        """
        return self.nbins_frame - self.skip_bins_start - self.skip_bins_end

    def to_waterfall(self, raw_data: np.ndarray) -> np.ndarray:
        """Function to convert from raw data to waterfall for use in
        reconstruction. The waterfall wil be shaped based on the
        settings found in the :class:`Frames` object.

        Shape of waterfall is generated based on parameters of the Frames
        object.

        Parameters
        ----------
        raw_data: ndarray
            1D array holding all measured raw data.
            The function works on a copy of the given raw_data.

        Returns
        -------
        waterfall: ndarray
            Raw-data shaped as waterfall (nprofiles, nbins).

        Raises
        ------
        RawDataImportError: Exception
            Raised if something is wrong with the raw data.
        """
        waterfall = self._assert_raw_data(raw_data)
        asrt.assert_frame_inputs(self)

        # Reshaping from raw data to waterfall
        waterfall = waterfall.reshape((self.nframes, self.nbins_frame))
        # Skips frames
        waterfall = waterfall[self.skip_frames:]

        # Skips bins at start and end of time frames.
        if self.skip_bins_end > 0:
            waterfall = waterfall[:, self.skip_bins_start:-self.skip_bins_end]
        else:
            waterfall = waterfall[:, self.skip_bins_start:]
        return waterfall

    # Check that provided raw data is valid.
    def _assert_raw_data(self, raw_data: t.Union[np.ndarray, t.Collection]) \
            -> np.ndarray:
        if not hasattr(raw_data, '__iter__'):
            raise expt.RawDataImportError('Raw data should be iterable')

        ndata = self.nframes * self.nbins_frame
        if not len(raw_data) == ndata:
            raise expt.RawDataImportError(
                f'Raw data has length {len(raw_data)}.\n'
                f'expected length: {ndata}')

        return np.array(raw_data)


def get_user_input(input: str = ''):
    """Function to let user provide input as a text file either as
    file path or as stdin.

    If a filepath is given through the **system arguments**, the path to the
    output directory can be given as the second argument. The specified
    output path will substitute the output path read from the input file.
    A failed path to the measured data, or the measured data itself can
    be given in the input file.

    If the input is provided via **stdin**, the measured data must be
    stored in the same input file.

    Returns
    -------
    tomo_input: tuple (list, ndarray)
        Tuple holding machine and reconstruction parameters as a list of
        strings directly read from text file and ndarray containing
        raw data (parameters, raw data).
    """
    if input == 'stdin' or input == '':
        read = _get_input_stdin()
    else:
        read = _get_input_args(input)
    return _split_input(read)


# Receive path to input file via sys.argv.
# Can also receive the path to the output directory.
def _get_input_args(input_file_pth: str) -> np.ndarray:
    if not os.path.isfile(input_file_pth):
        raise expt.InputError(f'The input file: "{input_file_pth}" '
                              f'does not exist!')

    with open(input_file_pth, 'r') as f:
        read = f.readlines()

    return np.array(read)


# Read machine parameters via stdin.
# Here the measured data must be pipelined in the same file as
# the machine parameters.
def _get_input_stdin() -> np.ndarray:
    read = []
    finished = False
    piped_raw_data = False

    line_num = 0
    ndata_points = PARAMETER_LENGTH

    while line_num < ndata_points:
        read.append(sys.stdin.readline())
        if line_num == RAW_DATA_FILE_IDX:
            if 'pipe' in read[-1]:
                piped_raw_data = True
        if piped_raw_data:
            if line_num == 16:
                nframes = int(read[-1])
            if line_num == 20:
                nbins = int(read[-1])
                ndata_points += nframes * nbins
        if line_num == ndata_points:
            finished = True
        line_num += 1

    return np.array(read)


# Splits the read input data to machine parameters and raw data.
# If the raw data is not already read from the input file, the
#  data will be found in the file given by the parameter file.
def _split_input(read_input: t.Sequence) -> t.Tuple[t.List, np.ndarray]:
    nframes_idx = 16
    nbins_idx = 20
    ndata = 0
    read_parameters = None
    read_data = None

    try:
        read_parameters = read_input[:PARAMETER_LENGTH]
        ndata = (int(read_parameters[nbins_idx])
                 * int(read_parameters[nframes_idx]))
        for i in range(PARAMETER_LENGTH):
            read_parameters[i] = read_parameters[i].strip('\r\n')
    except TypeError:
        err_msg = 'Something went wrong while accessing machine parameters.'
        raise expt.InputError(err_msg)

    if read_parameters[RAW_DATA_FILE_IDX] == 'pipe':
        try:
            read_data = np.array(read_input[PARAMETER_LENGTH:], dtype=float)
        except TypeError:
            err_msg = 'Pipelined raw-data could not be casted to float.'
            raise expt.InputError(err_msg)
    else:
        try:
            read_data = np.genfromtxt(read_parameters[RAW_DATA_FILE_IDX],
                                      dtype=float)
        except FileNotFoundError:
            err_msg = f'The given file path for the raw-data:\n' \
                      f'{read_parameters[RAW_DATA_FILE_IDX]}\n' \
                      f'Could not be found'
            raise FileNotFoundError(err_msg)
        except Exception:
            err_msg = 'Something went wrong while loading raw_data.'
            raise Exception(err_msg)

    if not len(read_data) == ndata:
        raise expt.InputError(f'Wrong amount of data points loaded.\n'
                              f'Expected: {ndata}\n'
                              f'Loaded:   {len(read_data)}')

    return read_parameters, read_data


def txt_input_to_machine(input_array: t.List) -> t.Tuple[Machine, Frames]:
    """Function converts the content of an input file and uses this to
    generate an machine object. The input file is given as a list
    holding one line of the file in ach element. The list should
    contain a direct read from a Fortran styled input file, where all
    lines should be included.

    Parameters
    ----------
    input_array: list
        Array containing each line of an input file as an element of the array.
        This should be a direct read from a tomography text file.

    Returns
    -------
    machine: Machine
        Machine object containing parameters for the machine and the
        tomographic reconstruction.
    frame: Frames
        TODO: A frame object

    Raises
    ------
    InputError: Exception
        Error in input array
    """

    # Checks if input is valid
    if not hasattr(input_array, '__iter__'):
        raise expt.InputError('Input should be iterable')
    if len(input_array) != PARAMETER_LENGTH:
        raise expt.InputError(f'Input array be of length {PARAMETER_LENGTH}, '
                              f'containing every line of input file '
                              f'(not including raw-data).')

    # String formatting
    for i in range(len(input_array)):
        input_array[i] = input_array[i].strip('\r\n')

    # Arguments for shaping waterfall from raw data
    fargs = {
        'raw_data_path': input_array[12],
        'framecount': int(input_array[16]),
        'skip_frames': int(input_array[18]),
        'framelength': int(input_array[20]),
        'dtbin': float(input_array[22]),
        'skip_bins_start': int(input_array[26]),
        'skip_bins_end': int(input_array[28]),
        'rebin': int(input_array[36])
    }

    # Creating waterfall from raw data
    frame = Frames(**fargs)
    nprofiles = frame.nprofs()
    nbins = frame.nbins()

    # Calculating limits in phase for reconstruction area
    # Specified by user.
    min_dt, max_dt = _min_max_dt(nbins, input_array)

    # Machine arguments
    # Some of the arguments are subtracted by one. This is
    # to convert from Fortran to C indexing.
    machine_kwargs = {
        'output_dir': input_array[14],
        'dtbin': float(input_array[22]),
        'dturns': int(input_array[24]),
        'synch_part_x': float(input_array[39]),
        'demax': float(input_array[41]),
        'filmstart': int(input_array[43]) - 1,
        'filmstop': int(input_array[45]) - 1,
        'filmstep': int(input_array[47]),
        'niter': int(input_array[49]),
        'snpt': int(input_array[51]),
        'full_pp_flag': bool(int(input_array[53])),
        'beam_ref_frame': int(input_array[55]) - 1,
        'machine_ref_frame': int(input_array[57]) - 1,
        'vrf1': float(input_array[61]),
        'vrf1dot': float(input_array[63]),
        'vrf2': float(input_array[65]),
        'vrf2dot': float(input_array[67]),
        'h_num': float(input_array[69]),
        'h_ratio': float(input_array[71]),
        'phi12': float(input_array[73]),
        'b0': float(input_array[75]),
        'bdot': float(input_array[77]),
        'mean_orbit_rad': float(input_array[79]),
        'bending_rad': float(input_array[81]),
        'trans_gamma': float(input_array[83]),
        'rest_energy': float(input_array[85]),
        'charge': float(input_array[87]),
        'self_field_flag': bool(int(input_array[91])),
        'g_coupling': float(input_array[93]),
        'zwall_over_n': float(input_array[95]),
        'pickup_sensitivity': float(input_array[97]),
        'nprofiles': nprofiles,
        'nbins': nbins,
        'min_dt': min_dt,
        'max_dt': max_dt
    }

    # Creating machine object
    machine = Machine(**machine_kwargs)

    return machine, frame


# Convert from setting min and max phase of reconstruction area
# as phase space coordinates to physical units of phase [s].
def _min_max_dt(nbins: int, input_array: t.Sequence) -> t.Tuple[float, float]:
    dtbin = float(input_array[22])
    min_dt_bin = int(input_array[31])
    max_dt_bin = int(input_array[34])

    min_dt = min_dt_bin * dtbin
    max_dt = (nbins - max_dt_bin) * dtbin
    return min_dt, max_dt


def raw_data_to_profiles(waterfall: np.ndarray, machine: MachineABC, rbn: int,
                         sampling_time: float,
                         synch_part_x: float = None) -> Profiles:
    """Converts from waterfall of untreated data, to waterfall
    ready for for tomography. The input waterfall is copied, and the
    treated waterfall is saved to a profiles object.

    The data treatment in this function is the same as in the original
    original tomography program:

    - Subtracts baseline from measurements
    - Re-binned profiles
    - Creates Profiles object to store waterfall
    - Negative values of waterfall is set to zero.

    **NB: dtbin and synch_part_x of the provided machine object
    will be updated after re-binning.**

    Parameters
    ----------
    waterfall: ndarray
        Raw-data shaped as waterfall (nprofiles, nbins).
    machine: Machine
        Machine object holding machine and reconstruction parameters.
    rbn: int
        Rebinning factor, the number of frame bins to
        rebin into one profile bin.
    sampling_time: float
        Size of profile bins [s].
    synch_part_x: float, optional, default=None
        X-coordinate of synchronous particle. If not given as argument,
        the machine.synch_part_x will be used.

    Returns
    -------
    profile: Profiles
        Profiles object holding the waterfall and information about the
        measurements.
    """
    if not hasattr(waterfall, '__iter__'):
        raise expt.WaterfallError('Waterfall should be an iterable')
    if synch_part_x is None:
        synch_part_x = machine.synch_part_x

    waterfall = np.array(waterfall)
    # Subtracting baseline
    waterfall[:] -= fortran.calc_baseline(waterfall, machine.beam_ref_frame)
    # Rebinning
    (waterfall,
     machine.dtbin,
     machine.synch_part_x) = pre_process.rebin(
        waterfall, rbn, sampling_time, synch_part_x)
    # Returning Profiles object.
    return Profiles(machine, sampling_time, waterfall)



"""Module containing the reconstruction algorithm derived from the cpp functions.

:Author(s): **Bernardo Abreu Figueiredo**
"""

import numpy as np
import cupy as cp
import logging

log = logging.getLogger(__name__)

def back_project(weights: cp.ndarray,
                 flat_points: cp.ndarray,
                 flat_profiles: cp.ndarray,
                 n_particles: int,
                 n_profiles: int) -> cp.ndarray:
    return cp.sum(cp.take(flat_profiles, flat_points, axis=0), axis=1) + weights

def project(flat_rec: cp.ndarray,
            flat_points: cp.ndarray,
            weights: cp.ndarray, n_particles: int,
            n_profiles: int, n_bins: int) -> cp.ndarray:
    return cp.bincount(flat_points.reshape(-1), weights.repeat(n_profiles), minlength=n_profiles * n_bins) + flat_rec

def normalize(flat_rec: cp.ndarray,
              n_profiles: int, n_bins: int) -> cp.ndarray:

    flat_rec = flat_rec.reshape((n_profiles, n_bins))
    sum_profile = cp.sum(flat_rec, axis=1)
    flat_rec /= cp.expand_dims(sum_profile, axis=1)
    flat_rec = flat_rec.reshape(-1)
    sum_waterfall = cp.sum(sum_profile)

    if sum_waterfall <= 0:
        raise RuntimeError("Phase space reduced to zeros!")
    return flat_rec

def clip(array: cp.ndarray,
        clip_val: float) -> cp.ndarray:
    array[array < clip_val] = clip_val
    return array

def find_difference_profile(flat_rec: cp.ndarray,
                            flat_profiles: cp.ndarray) -> cp.ndarray:
    return flat_profiles - flat_rec

def discrepancy(diff_prof: cp.ndarray,
                n_profiles: int, n_bins: int) -> float:
    all_bins = n_profiles * n_bins
    squared_sum = cp.sum(cp.power(diff_prof, 2))

    return cp.sqrt(squared_sum / all_bins)

def compensate_particle_amount(diff_prof: cp.ndarray,
                               rparts: cp.ndarray,
                               n_profiles: int, n_bins: int) -> cp.ndarray:
    return diff_prof * rparts.reshape(n_profiles * n_bins)

def max_2d(array: cp.ndarray,
           x_axis: int, y_axis: int) -> float:
    return cp.max(array[:y_axis, :x_axis])

def count_particles_in_bin(xp: cp.ndarray,
                           n_profiles: int, n_particles: int,
                           n_bins: int) -> cp.ndarray:

    bins = cp.arange(n_bins + 1)
    rparts = cp.empty((n_profiles, n_bins), dtype=cp.int32)
    for j in range(n_profiles):
        rparts[j], _ = cp.histogram(xp[:, j], bins=bins)
    return rparts

def reciprocal_particles(xp: cp.ndarray,
                         n_bins: int, n_profiles: int,
                         n_particles: int) -> cp.ndarray:

    rparts = count_particles_in_bin(xp, n_profiles, n_particles, n_bins)
    max_bin_val = max_2d(rparts, n_particles, n_profiles)

    # Setting 0's to 1's to avoid zero division
    rparts = cp.where(rparts == 0, 1, rparts)
    rparts = max_bin_val / rparts

    return rparts

def create_flat_points(xp: cp.ndarray,
                       n_particles: int, n_profiles: int,
                       n_bins: int) -> cp.ndarray:
    flat_points = cp.copy(xp)
    flat_points += cp.arange(n_profiles) * n_bins

    return flat_points

def reconstruct_cupy(xp: cp.ndarray,
                waterfall: cp.ndarray, n_iter: int,
                n_bins: int, n_particles: int, n_profiles: int,
                verbose: bool = False) -> tuple:

    weights = cp.zeros(n_particles)
    discr = np.zeros(n_iter + 1)
    flat_profiles = waterfall.flatten()
    flat_rec = cp.zeros(n_profiles * n_bins)

    all_bins = n_profiles * n_bins
    diff_prof = cp.zeros(all_bins)
    flat_points = cp.zeros(n_particles * n_profiles)

    # Actual functionality
    rparts = reciprocal_particles(xp, n_bins, n_profiles, n_particles)
    flat_points = create_flat_points(xp, n_particles, n_profiles, n_bins)
    weights = back_project(weights, flat_points, flat_profiles, n_particles, n_profiles)
    weights = clip(weights, 0.0)

    if cp.sum(weights) <= 0:
        raise RuntimeError("All of phase space got reduced to zeros")

    if verbose:
        print(" Iterating...")

    for iteration in range(n_iter):
        if verbose:
            print(f"{iteration+1:3}")

        flat_rec = project(flat_rec, flat_points, weights, n_particles, n_profiles, n_bins)
        flat_rec = normalize(flat_rec, n_profiles, n_bins)
        diff_prof = find_difference_profile(flat_rec, flat_profiles)
        discr[iteration] = discrepancy(diff_prof, n_profiles, n_bins)
        diff_prof = compensate_particle_amount(diff_prof, rparts, n_profiles, n_bins)
        weights = back_project(weights, flat_points, diff_prof, n_particles, n_profiles)
        weights = clip(weights, 0.0)


        if cp.sum(weights) <= 0:
            raise RuntimeError("All of phase space got reduced to zeros")

    # Calculating final discrepancy
    flat_rec = project(flat_rec, flat_points, weights, n_particles, n_profiles, n_bins)
    flat_rec = normalize(flat_rec, n_profiles, n_bins)
    diff_prof = find_difference_profile(flat_rec, flat_profiles)
    discr[n_iter] = discrepancy(diff_prof, n_profiles, n_bins)

    if verbose:
        print("Done!")

    return weights, discr, flat_rec


"""Unit-tests for the Machine class.

Run as python test_machine.py in console or via coverage
"""
from __future__ import annotations
import unittest

from .. import commons
import longitudinal_tomography.tracking.machine as mch


# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestMachine(unittest.TestCase):

    def test_set_nbins_correct_nbins(self):
        new_nbins = 200
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nbins = new_nbins
        self.assertEqual(machine.nbins, new_nbins,
                         msg='nbins was set incorrectly')

    def test_set_nbins_correct_yat0(self):
        new_nbins = 200
        machine = mch.Machine(**MACHINE_ARGS)
        machine.nbins = new_nbins
        self.assertEqual(machine.synch_part_y, 100.0,
                         msg='yat0 was set incorrectly when nbins was updated')

    def test_values_at_turns_correct_length(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        self.assertEqual(len(machine.time_at_turn), 746,
                         msg='Wrong length of array: time_at_turn')
        self.assertEqual(len(machine.omega_rev0), 746,
                         msg='Wrong length of array: omega_rev0')
        self.assertEqual(len(machine.phi0), 746,
                         msg='Wrong length of array: phi0')
        self.assertEqual(len(machine.drift_coef), 746,
                         msg='Wrong length of array: drift_coef')
        self.assertEqual(len(machine.deltaE0), 746,
                         msg='Wrong length of array: deltaE0')
        self.assertEqual(len(machine.beta0), 746,
                         msg='Wrong length of array: beta0')
        self.assertEqual(len(machine.eta0), 746,
                         msg='Wrong length of array: eta0')
        self.assertEqual(len(machine.e0), 746,
                         msg='Wrong length of array: e0')
        self.assertEqual(len(machine.vrf1_at_turn), 746,
                         msg='Wrong length of array: vrf1_at_turn')
        self.assertEqual(len(machine.vrf2_at_turn), 746,
                         msg='Wrong length of array: vrf2_at_turn')

    def test_values_at_turns_correct_time_at_turn(self):
        nprofs = 5
        machine_ref_frame = 2

        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = nprofs
        machine.machine_ref_frame = machine_ref_frame
        machine.values_at_turns()

        correct = [-1.37016417e-05, -1.23314411e-05, -1.09612485e-05,
                   -9.59106409e-06, -8.22088776e-06, -6.85071954e-06,
                   -5.48055942e-06, -4.11040741e-06, -2.74026350e-06,
                   -1.37012770e-06, 0.00000000e+00, 1.37011959e-06,
                   2.74023109e-06, 4.11033447e-06, 5.48042976e-06,
                   6.85051693e-06, 8.22059601e-06, 9.59066698e-06,
                   1.09607299e-05, 1.23307846e-05, 1.37008313e-05]

        for tat, corr in zip(machine.time_at_turn, correct):
            self.assertAlmostEqual(tat, corr,
                                   msg='Error in calculation of time at turn')

    def test_values_at_turns_correct_omega_rev0(self):
        nprofs = 5
        machine_ref_frame = 2

        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = nprofs
        machine.machine_ref_frame = machine_ref_frame
        machine.values_at_turns()

        correct = [4585595.05834437, 4585622.18581794, 4585649.31304857,
                   4585676.44003627, 4585703.56678104, 4585730.69328289,
                   4585757.81954182, 4585784.94555784, 4585812.07133096,
                   4585839.19686116, 4585866.32214847, 4585893.44719289,
                   4585920.57199441, 4585947.69655305, 4585974.82086881,
                   4586001.94494169, 4586029.0687717,  4586056.19235885,
                   4586083.31570313, 4586110.43880456, 4586137.56166313]

        for omega, corr in zip(machine.omega_rev0, correct):
            self.assertAlmostEqual(omega, corr,
                                   msg='Error in calculation of revolution '
                                       'frequency (omega_rev0)')

    def test_values_at_turns_correct_phi0(self):
        nprofs = 5
        machine_ref_frame = 2

        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = nprofs
        machine.machine_ref_frame = machine_ref_frame
        machine.values_at_turns()

        correct = [0.40078213, 0.40078213, 0.40078213, 0.40078213,
                   0.40078213, 0.40078213, 0.40078213, 0.40078213,
                   0.40078213, 0.40078213, 0.40078213, 0.40078213,
                   0.40078213, 0.40078213, 0.40078213, 0.40078213,
                   0.40078213, 0.40078213, 0.40078213, 0.40078213,
                   0.40078213]

        for phi, corr in zip(machine.phi0, correct):
            self.assertAlmostEqual(phi, corr,
                                   msg='Error in calculation of synchronous '
                                       'phase at each turn (phi0)')

    def test_values_at_turns_correct_drift_coef(self):
        nprofs = 5
        machine_ref_frame = 2

        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = nprofs
        machine.machine_ref_frame = machine_ref_frame
        machine.values_at_turns()

        correct = [3.36099331e-08, 3.36094281e-08, 3.36089232e-08,
                   3.36084183e-08, 3.36079135e-08, 3.36074086e-08,
                   3.36069038e-08, 3.36063989e-08, 3.36058941e-08,
                   3.36053893e-08, 3.36048845e-08, 3.36043797e-08,
                   3.36038749e-08, 3.36033702e-08, 3.36028654e-08,
                   3.36023607e-08, 3.36018560e-08, 3.36013513e-08,
                   3.36008466e-08, 3.36003419e-08, 3.35998373e-08]

        for drift, corr in zip(machine.drift_coef, correct):
            self.assertAlmostEqual(drift, corr,
                                   msg='Error in calculation of drift '
                                       'coefficient (drift_coef)')

    def test_values_at_turns_correct_deltaE0(self):
        nprofs = 5
        machine_ref_frame = 2

        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = nprofs
        machine.machine_ref_frame = machine_ref_frame
        machine.values_at_turns()

        correct = [1028.87237942, 1028.87237942, 1028.87237942, 1028.87237942,
                   1028.87237942, 1028.87237942, 1028.87237942, 1028.87237942,
                   1028.87237942, 1028.87237942,    0.,         1028.87237942,
                   1028.87237942, 1028.87237942, 1028.87237942, 1028.87237942,
                   1028.87237942, 1028.87237942, 1028.87237942, 1028.87237942,
                   1028.87237942]

        for dE0, corr in zip(machine.deltaE0, correct):
            self.assertAlmostEqual(dE0, corr,
                                   msg='Error in calculation of energy '
                                       'difference of synch part pr turn '
                                       '(deltaE0)')

    def test_values_at_turns_correct_beta0(self):
        nprofs = 5
        machine_ref_frame = 2

        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = nprofs
        machine.machine_ref_frame = machine_ref_frame
        machine.values_at_turns()

        correct = [0.38239747, 0.38239973, 0.38240199, 0.38240425, 0.38240652,
                   0.38240878, 0.38241104, 0.3824133,  0.38241556, 0.38241783,
                   0.38242009, 0.38242235, 0.38242461, 0.38242687, 0.38242914,
                   0.3824314,  0.38243366, 0.38243592, 0.38243818, 0.38244044,
                   0.38244271]

        for beta, corr in zip(machine.beta0, correct):
            self.assertAlmostEqual(beta, corr,
                                   msg='Error in calculation of relativistic '
                                       'beta (beta0)')

    def test_values_at_turns_correct_eta0(self):
        nprofs = 5
        machine_ref_frame = 2

        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = nprofs
        machine.machine_ref_frame = machine_ref_frame
        machine.values_at_turns()

        correct = [0.79428378, 0.79428205, 0.79428032, 0.79427859, 0.79427686,
                   0.79427513, 0.7942734,  0.79427167, 0.79426994, 0.79426821,
                   0.79426648, 0.79426475, 0.79426302, 0.79426129, 0.79425956,
                   0.79425783, 0.7942561,  0.79425437, 0.79425264, 0.79425091,
                   0.79424918]

        for eta, corr in zip(machine.eta0, correct):
            self.assertAlmostEqual(eta, corr,
                                   msg='Error in calculation of phase slip '
                                       'factor (eta0)')

    # This array is tested as integers due to its high values.
    def test_values_at_turns_correct_e0(self):
        nprofs = 5
        machine_ref_frame = 2

        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = nprofs
        machine.machine_ref_frame = machine_ref_frame
        machine.values_at_turns()

        correct = [1015448496, 1015449524, 1015450553, 1015451582, 1015452611,
                   1015453640, 1015454669, 1015455698, 1015456727, 1015457755,
                   1015458784, 1015459813, 1015460842, 1015461871, 1015462900,
                   1015463929, 1015464958, 1015465986, 1015467015, 1015468044,
                   1015469073]

        for e0, corr in zip(machine.e0, correct):
            self.assertEqual(int(e0), corr,
                             msg='Error in calculation of energy '
                                 'of synch. particle (e0)')

    def test_values_at_turns_correct_vrf_with_derivative(self):
        nprofs = 5
        machine_ref_frame = 2

        machine = mch.Machine(**MACHINE_ARGS)
        machine.nprofiles = nprofs
        machine.machine_ref_frame = machine_ref_frame
        machine.vrf1dot = 10
        machine.values_at_turns()

        correct = [2637.19689392, 2637.19690762, 2637.19692132, 2637.19693502,
                   2637.19694872, 2637.19696243, 2637.19697613, 2637.19698983,
                   2637.19700353, 2637.19701723, 2637.19703093, 2637.19704463,
                   2637.19705834, 2637.19707204, 2637.19708574, 2637.19709944,
                   2637.19711314, 2637.19712684, 2637.19714054, 2637.19715424,
                   2637.19716794]

        for vrf, corr in zip(machine.vrf1_at_turn, correct):
            self.assertAlmostEqual(vrf, corr,
                                   msg='Error in calculation of RF voltage '
                                       'with vrf1dot (vrf1_at_turn)')



import logging as log
import os
import time as tm

import matplotlib.pyplot as plt
import numpy as np

import longitudinal_tomography.tomography.tomography as tomography
import longitudinal_tomography.tracking.particles as parts
import longitudinal_tomography.tracking.tracking as tracking
import longitudinal_tomography.utils.tomo_input as tomoin

import longitudinal_tomography
longitudinal_tomography.use_gpu()

file_name = 'C550MidPhaseNoise.dat'
ex_dir = os.path.split(os.path.realpath(os.path.dirname(__file__)))[0]
in_file_pth = os.path.join(ex_dir, 'input_files', file_name)

parameter_lines = 98
input_parameters = []
with open(in_file_pth, 'r') as line:
    for i in range(parameter_lines):
        input_parameters.append(line.readline().strip())

raw_data = np.genfromtxt(in_file_pth, skip_header=98)

machine, frames = tomoin.txt_input_to_machine(input_parameters)
machine.values_at_turns()

measured_waterfall = frames.to_waterfall(raw_data)

profiles = tomoin.raw_data_to_profiles(
    measured_waterfall, machine,
    frames.rebin, frames.sampling_time)

tomo = tomography.TomographyCpp(profiles.waterfall)

snpt0 = 1
snpt1 = 6
dsnpt = 1
snpts = np.arange(snpt0, snpt1, dsnpt)

diffs = []
dtimes = []
successful_inputs = []
for snpt in snpts:
    print(f'Running tomo using {snpt ** 2} particles per cell of phase space.')
    machine.snpt = snpt

    try:
        t0 = tm.perf_counter()

        tracker = tracking.Tracking(machine)
        xp, yp = tracker.track(machine.filmstart)

        xp, yp = parts.physical_to_coords(
            xp, yp, machine, tracker.particles.xorigin,
            tracker.particles.dEbin)
        xp, yp = parts.ready_for_tomography(xp, yp, machine.nbins)

        tomo.xp = xp
        weight = tomo.run(niter=machine.niter)
        tomo.xp = None

        t1 = tm.perf_counter()
    except Exception as e:
        log.warning(f' {file_name} could not be run with '
                    f'input parameter snpt = {snpt}.\n'
                    f'Error message: {e}')
    else:
        dtimes.append(t1 - t0)
        diffs.append(tomo.diff[-1])
        successful_inputs.append(snpt)

# Plotting
if len(successful_inputs) > 0:
    for value, discr, time in zip(successful_inputs, diffs, dtimes):
        plt.scatter(time, discr, label=f'{value}')

    offset = max(diffs) * 0.05
    ylim_up = max(diffs) + offset
    ylim_low = min(diffs) - offset
    plt.ylim([ylim_low, ylim_up])

    ax = plt.gca()
    ax.legend(title='snpt')
    ax.set_xlabel('Execution time [s]')
    ax.set_ylabel('Discrepancy')
    ax.ticklabel_format(axis='y', scilimits=(0, 0), style='sci')
    plt.show()
else:
    print('No runs were successful.')



import os

import numpy as np

import longitudinal_tomography.tracking.machine as mch
import longitudinal_tomography.tomography.tomography as tomography
import longitudinal_tomography.tracking.particles as parts
import longitudinal_tomography.tracking.tracking as tracking
import longitudinal_tomography.utils.tomo_input as tomoin
import longitudinal_tomography.utils.tomo_output as tomoout

import longitudinal_tomography
longitudinal_tomography.use_gpu()


ex_dir = os.path.split(os.path.realpath(os.path.dirname(__file__)))[0]
in_file_pth = os.path.join(ex_dir, 'input_files', 'INDIVShavingC325.dat')

# All values retrieved from INDIVShavingC325.dat
dtbin = 9.999999999999999E-10

# This class is here to help shaping your waterfall from a fortran style
#  collection and format of measured raw data.
# This can be omitted by shaping the waterfall yourself.
frame_input_args = {
    'raw_data_path':    in_file_pth,
    'framecount':       150,
    'skip_frames':      0,
    'framelength':      1000,
    'dtbin':            dtbin,
    'skip_bins_start':  170,
    'skip_bins_end':    70,
    'rebin':            3
}

frames = tomoin.Frames(**frame_input_args)
nprofiles = frames.nprofs()
nbins = frames.nbins()

# Machine and reconstruction parameters
machine_args = {
    'output_dir':           '/tmp/',
    'dtbin':                dtbin,
    'dturns':               5,
    'synch_part_x':         334.00000000000006,
    'demax':                -1.E6,              # noqa
    'filmstart':            0,
    'filmstop':             1,
    'filmstep':             1,
    'niter':                20,
    'snpt':                 4,
    'full_pp_flag':         False,
    'beam_ref_frame':       0,
    'machine_ref_frame':    0,
    'vrf1':                 2637.197030932989,
    'vrf1dot':              0.0,
    'vrf2':                 0.0,
    'vrf2dot':              0.0,
    'h_num':                1,
    'h_ratio':              2.0,
    'phi12':                0.4007821253666541,
    'b0':                   0.15722,
    'bdot':                 0.7949999999999925,
    'mean_orbit_rad':       25.0,
    'bending_rad':          8.239,
    'trans_gamma':          4.1,
    'rest_energy':          0.93827231E9,
    'charge':               1,
    'self_field_flag':      False,
    'g_coupling':           0.0,
    'zwall_over_n':         0.0,
    'pickup_sensitivity':   0.36,
    'nprofiles':            nprofiles,
    'nbins':                nbins,
    'min_dt':               0.0,
    'max_dt':               dtbin * nbins
}

machine = mch.Machine(**machine_args)

raw_data = np.genfromtxt(frames.raw_data_path, skip_header=98,
                         dtype=np.float32)
machine.values_at_turns()
measured_waterfall = frames.to_waterfall(raw_data)

if machine.synch_part_x < 0 or machine.self_field_flag:
    msg = 'This example does not include fitting or self field tracking.'
    raise NotImplementedError(msg)

profiles = tomoin.raw_data_to_profiles(
    measured_waterfall, machine,
    frames.rebin, frames.sampling_time)

tracker = tracking.Tracking(machine)
reconstruct_idx = machine.filmstart

# profile charge needed for fortran style output during tracking.
profiles.calc_profilecharge()
xp, yp = tracker.track(reconstruct_idx)

# Converts from physical units to phase space coordinates as bin numbers.
xp, yp = parts.physical_to_coords(
    xp, yp, machine, tracker.particles.xorigin,
    tracker.particles.dEbin)

# Filters out lost particles, transposes particle matrix, casts to np.int32.
xp, yp = parts.ready_for_tomography(xp, yp, machine.nbins)

# Reconstructing phase space
tomo = tomography.TomographyCpp(profiles.waterfall, xp)
weight = tomo.run(niter=machine.niter)

# Creating image for fortran style presentation of phase space.
image = tomoout.create_phase_space_image(
    xp, yp, weight, machine.nbins, reconstruct_idx)
tomoout.show(image, tomo.diff, profiles.waterfall[reconstruct_idx])



"""Unit-tests for the assertions module.

Run as python test_assertions.py in console or via coverage
"""
from __future__ import annotations
import unittest

import numpy as np

from . import commons
import longitudinal_tomography.tracking.machine as mch
from longitudinal_tomography import assertions as asrt, exceptions as expt
import longitudinal_tomography.utils.tomo_input as tomoin


MACHINE_ARGS = commons.get_machine_args()


class TestAssertions(unittest.TestCase):

    def test_assert_greater(self):
        var_name = 'test'
        limit = 15
        error_class = AssertionError

        var = 10
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_greater(var, var_name, limit, error_class)

        var = limit
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_greater(var, var_name, limit, error_class)

        var = 20
        asrt.assert_greater(var, var_name, limit, error_class)

    def test_assert_less(self):
        var_name = 'test'
        limit = 15
        error_class = AssertionError

        var = 20
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_less(var, var_name, limit, error_class)

        var = limit
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_less(var, var_name, limit, error_class)

        var = 10
        asrt.assert_less(var, var_name, limit, error_class)

    def test_assert_equal(self):
        var_name = 'test'
        limit = 15
        error_class = AssertionError

        var = 10
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_equal(var, var_name, limit, error_class)

        var = 20
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_equal(var, var_name, limit, error_class)

        var = limit
        asrt.assert_equal(var, var_name, limit, error_class)

    def test_assert_not_equal(self):
        var_name = 'test'
        limit = 15
        error_class = AssertionError

        var = 10
        asrt.assert_not_equal(var, var_name, limit, error_class)

        var = limit
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_not_equal(var, var_name, limit, error_class)

        var = 20
        asrt.assert_not_equal(var, var_name, limit, error_class)

    def test_assert_less_or_equal(self):
        var_name = 'test'
        limit = 15
        error_class = AssertionError

        var = 10
        asrt.assert_less_or_equal(var, var_name, limit, error_class)

        var = limit
        asrt.assert_less_or_equal(var, var_name, limit, error_class)

        var = 20
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_less_or_equal(var, var_name, limit, error_class)

    def test_assert_greater_or_equal(self):
        var_name = 'test'
        limit = 15
        error_class = AssertionError

        var = 10
        asrt.assert_less_or_equal(var, var_name, limit, error_class)

        var = limit
        asrt.assert_less_or_equal(var, var_name, limit, error_class)

        var = 20
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_less_or_equal(var, var_name, limit, error_class)

    def test_assert_inrange(self):
        var_name = 'test'
        low_lim = 15
        up_lim = 20
        error_class = AssertionError

        var = low_lim
        asrt.assert_inrange(var, var_name, low_lim, up_lim, error_class)

        var = up_lim
        asrt.assert_inrange(var, var_name, low_lim, up_lim, error_class)

        var = 17
        asrt.assert_inrange(var, var_name, low_lim, up_lim, error_class)

        var = 10
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_inrange(var, var_name, low_lim, up_lim, error_class)

        var = 25
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_inrange(var, var_name, low_lim, up_lim, error_class)

    def test_assert_array_not_equal(self):
        array = np.ones((10, 10))
        array_name = 'test_array'
        limit = 1
        error_class = AssertionError

        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_array_not_equal(array, array_name, limit, error_class)

        array[1, 1] = 2
        asrt.assert_array_not_equal(array, array_name, limit, error_class)

    def test_assert_array_shape_equal(self):
        a_array = np.zeros((2, 4))
        b_array = np.zeros((2, 4))
        c_array = np.zeros((7, 4))
        correct_shape = (2, 4)

        with self.assertRaises(
                expt.UnequalArrayShapes,
                msg='An error should have been raised'):
            asrt.assert_array_shape_equal(
                (a_array, b_array, c_array),
                ('a', 'b', 'c'), correct_shape)

        asrt.assert_array_shape_equal((a_array, b_array),
                                      ('a', 'b'), correct_shape)

    def test_assert_array_in_range(self):
        low_lim = 5
        up_lim = 15
        error_class = AssertionError

        array = np.arange(1, 20)

        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_array_in_range(array, low_lim, up_lim, error_class)

        array = np.arange(low_lim, up_lim+1)
        asrt.assert_array_in_range(array, low_lim, up_lim, error_class)

    def test_assert_array_greater(self):
        limit = 10
        error_class = AssertionError

        array = np.ones(10) * 15
        asrt.assert_array_greater(array, limit, error_class)

        array[3] = limit
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_array_greater(array, limit, error_class)

        array[3] = 9
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_array_greater(array, limit, error_class)

    def test_assert_array_greater_or_equal(self):
        limit = 10
        error_class = AssertionError

        array = np.ones(10) * 15
        asrt.assert_array_greater_eq(array, limit, error_class)

        array[3] = limit
        asrt.assert_array_greater_eq(array, limit, error_class)

        array[3] = 9
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_array_greater_eq(array, limit, error_class)

    def test_assert_array_less(self):
        limit = 10
        error_class = AssertionError

        array = np.ones(10) * 5
        asrt.assert_array_less(array, limit, error_class)

        array[3] = limit
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_array_less(array, limit, error_class)

        array[3] = 11
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_array_less(array, limit, error_class)

    def test_assert_array_less_or_equal(self):
        limit = 10
        error_class = AssertionError

        array = np.ones(10) * 5
        asrt.assert_array_less_eq(array, limit, error_class)

        array[3] = limit
        asrt.assert_array_less_eq(array, limit, error_class)

        array[3] = 11
        with self.assertRaises(
                error_class, msg='An error should have been raised'):
            asrt.assert_array_less_eq(array, limit, error_class)

    def test_assert_only_valid_particles(self):
        nbins = 25
        particles = np.ones((10, 10))

        # Testing too large x coordinate
        particles[6, 9] = nbins
        with self.assertRaises(
                expt.InvalidParticleError,
                msg='Particles outside of the image width should raise an '
                    'exception.'):
            asrt.assert_only_valid_particles(particles, nbins)

        particles[6, 9] = -1
        with self.assertRaises(
                expt.InvalidParticleError,
                msg='Particles outside of the image width should raise an '
                    'exception.'):
            asrt.assert_only_valid_particles(particles, nbins)

    def test_assert_fields(self):
        error_class = AssertionError
        needed_fields = ['hope']
        machine = mch.Machine(**MACHINE_ARGS)

        with self.assertRaises(error_class,
                               msg='Object lacking needed fields '
                                   'should raise an exception'):
            asrt.assert_fields(
                machine, 'test_machine', needed_fields, error_class)

    def test_assert_machine_input_bad_dtbin(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.dtbin = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='dtbin with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.dtbin = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='dtbin with value of 0 should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_dturns(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.dturns = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='dturns with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.dturns = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='dturns with value of 0 should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_filmsss(self):
        machine = mch.Machine(**MACHINE_ARGS)

        # Testing filmstart
        machine.filmstart = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='filmstart with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.filmstart = -1
        with self.assertRaises(
                expt.MachineParameterError,
                msg='filmstart with negative value should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.filmstart = 2
        with self.assertRaises(
                expt.MachineParameterError,
                msg='filmstart with value larger than filmstop '
                    'should raise an Exception'):
            asrt.assert_machine_input(machine)
        machine.filmstart = 0

        # Testing filmstep

        machine.filmstep = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='filmstep with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.filmstep = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='filmstep with negative value should raise an Exception'):
            asrt.assert_machine_input(machine)
        machine.filmstep = 1

        # Testing filmstop

        machine.filmstop = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='filmstop with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_niter(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.niter = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='niter with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.niter = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='niter with value less than 1 should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_snpt(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.snpt = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='snpt with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.snpt = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='snpt with value less than 1 should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_ref_frames(self):
        machine = mch.Machine(**MACHINE_ARGS)

        # Testing machine reference frame
        machine.machine_ref_frame = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='machine_ref_frame with value None should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.machine_ref_frame = -1
        with self.assertRaises(
                expt.MachineParameterError,
                msg='machine_ref_frame with value less than zero should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)
        machine.machine_ref_frame = 1

        # Testing beam reference frame
        machine.beam_ref_frame = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='beam_ref_frame with value None should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.beam_ref_frame = -1
        with self.assertRaises(
                expt.MachineParameterError,
                msg='beam_ref_frame with value less than zero should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_hnum(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.h_num = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='h_num with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.h_num = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='h_num with value less than one should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_h_ratio(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.h_ratio = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='h_ratio with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.h_ratio = -1
        with self.assertRaises(
                expt.MachineParameterError,
                msg='h_ratio with value less than zero should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_b0(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.b0 = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='b0 with value None should raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.b0 = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='b0 with with value less than or equal to zero should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_mean_orbit_rad(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.mean_orbit_rad = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='mean_orbit_rad with value None should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.mean_orbit_rad = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='mean_orbit_rad with with value less than or '
                    'equal to zero should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_bending_rad(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.bending_rad = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='bending_rad with value None should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.bending_rad = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='bending_rad with with value less than or '
                    'equal to zero should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_e_rest(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.e_rest = None
        with self.assertRaises(
                expt.MachineParameterError,
                msg='e_rest with value None should '
                    'raise an Exception'):
            asrt.assert_machine_input(machine)

        machine.e_rest = 0
        with self.assertRaises(
                expt.MachineParameterError,
                msg='e_rest with with value less than or '
                    'equal to zero should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_pickup_sensitivity(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.pickup_sensitivity = -1
        with self.assertRaises(
                expt.SpaceChargeParameterError,
                msg='pickup_sensitivity with with value less than '
                    'zero should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_machine_input_bad_g_coupling(self):
        machine = mch.Machine(**MACHINE_ARGS)

        machine.g_coupling = -1
        with self.assertRaises(
                expt.SpaceChargeParameterError,
                msg='g_coupling with with value less than '
                    'zero should raise an Exception'):
            asrt.assert_machine_input(machine)

    def test_assert_frame_input(self):
        frame = tomoin.Frames(
            framecount=10, framelength=10, skip_frames=2,
            skip_bins_start=2, skip_bins_end=2, rebin=1, dtbin=1e-7)

        # Checking nframes
        frame.nframes = 0
        with self.assertRaises(
                expt.InputError,
                msg='nframes equal to zero should raise an Exception'):
            asrt.assert_frame_inputs(frame)
        frame.nframes = 10

        # Checking skip_frames
        frame.skip_frames = -1
        with self.assertRaises(
                expt.InputError,
                msg='skip_frames with negative value '
                    'should raise an Exception'):
            asrt.assert_frame_inputs(frame)

        frame.skip_frames = frame.nframes
        with self.assertRaises(
                expt.InputError,
                msg='skip_frames with value equal to nframes '
                    'should raise an Exception'):
            asrt.assert_frame_inputs(frame)
        frame.skip_frames = 2

        # Checking nbins_frame
        frame.nbins_frame = -1
        with self.assertRaises(
                expt.InputError,
                msg='nbins_frame with value lower than zero '
                    'should raise an Exception'):
            asrt.assert_frame_inputs(frame)
        frame.nbins_frame = 10

        # Checking skip_bins_start
        frame.skip_bins_start = -1
        with self.assertRaises(
                expt.InputError,
                msg='skip_bins_start with value lower than zero '
                    'should raise an Exception'):
            asrt.assert_frame_inputs(frame)

        frame.skip_bins_start = frame.nbins_frame
        with self.assertRaises(
                expt.InputError,
                msg='skip_bins_start with value equal to nbins_frame '
                    'should raise an Exception'):
            asrt.assert_frame_inputs(frame)
        frame.skip_bins_start = 2

        # Checking skip_bins_end
        frame.skip_bins_end = -1
        with self.assertRaises(
                expt.InputError,
                msg='skip_bins_end with value lower than zero '
                    'should raise an Exception'):
            asrt.assert_frame_inputs(frame)

        frame.skip_bins_end = frame.nbins_frame
        with self.assertRaises(
                expt.InputError,
                msg='skip_bins_end with value equal to nbins_frame '
                    'should raise an Exception'):
            asrt.assert_frame_inputs(frame)
        frame.skip_bins_end = 2

        # Checking total skipped bins
        frame.skip_bins_start = 5
        frame.skip_bins_end = 5
        with self.assertRaises(
                expt.InputError,
                msg='total number of skipped bins greater than or equal to '
                    'nbins_frame should raise an Exception'):
            asrt.assert_frame_inputs(frame)

        frame.skip_bins_start = 2
        frame.skip_bins_end = 2

        frame.rebin = 0
        with self.assertRaises(
                expt.InputError,
                msg='skip_bins_start with value lower than zero '
                    'should raise an Exception'):
            asrt.assert_frame_inputs(frame)

    def test_assert_var_not_none(self):
        error_class = AssertionError
        with self.assertRaises(error_class,
                               msg='An exception should have been raised'):
            asrt.assert_var_not_none(None, 'test', error_class)



"""Module containing Machine class for storing
machine and reconstruction parameters

:Author(s): **Christoffer Hjertø Grindheim**, **Anton Lu**
"""
from __future__ import annotations
import logging
from typing import Tuple

import numpy as np
from scipy import optimize, constants

from .. import assertions as asrt
from ..utils import physics
from .machine_base import MachineABC

log = logging.getLogger(__name__)

_machine_opts_def = {
    'vrf1dot': 0.0,
    'vrf2': 0.0,
    'vrf2dot': 0.0,
    'bdot': 0.0,
}


class Machine(MachineABC):
    """Class holding machine and reconstruction parameters.

    This class holds machine parameters and information about the measurements.
    Also, it holds settings for the reconstruction process.

    The Machine class and its values are needed for the original particle
    tracking routine. Its values are used for calculation of reconstruction
    area and info concerning the phase space, the distribution of particles,
    and the tracking itself. In addition to this, the machine object is needed
    for the generation of
    :class:`~longitudinal_tomography.data.profiles.Profiles` objects.

    See superclass for documentation about inherited class variables.

    To summarize, the Machine class must be used if a program resembling the
    original Fortran version is to be created.

    Parameters
    ----------
    dturns: int
        Number of machine turns between each measurement.
    vrf1: float
        Peak voltage of the first RF system at the machine reference frame.
    mean_orbit_rad: float
        Mean orbit radius of machine [m].
    bending_rad: float
        Machine bending radius [m].
    b0: float
        B-field at machine reference frame [T].
    bdot: float
        Time derivative of B-field (considered constant) [T/s].
    trans_gamma: float
        Transitional gamma.
    e_rest: float
        Rest energy of accelerated particle [eV/C^2], saved as e_rest.
    nprofiles: int
        Number of measured profiles.
    nbins: int
        Number of bins in a profile.
    synch_part_x: float
        Synchronous phase given in number of bins, counting\
        from the lower profile bound to the synchronous phase.
    dtbin: float
        Size of profile bins [s].
    kwargs:
        All scalar attributes can be set via the kwargs.

    Attributes
    ----------
    vrf1: float
        Peak voltage of the first RF system at the machine reference frame.
    vrf2: float, default=0.0
        Peak voltage of the second RF system at the machine reference frame.\n
    vrf1dot: float, default=0.0
        Time derivatives of the voltages of the first RF system
        (considered constant).\n
    vrf2dot: float, default=0.0
        Time derivatives of the voltages of the second RF system
        (considered constant).\n
    b0: float
        B-field at machine reference frame [T].
    bdot: float
        Time derivative of B-field (considered constant) [T/s].
    h_num: int, default=1
        Principle harmonic number.\n
    """

    def __init__(self, dturns: int, vrf1: float, mean_orbit_rad: float,
                 bending_rad: float, b0: float, trans_gamma: float,
                 rest_energy: float, nprofiles: int, nbins: int, dtbin: float,
                 vat_now: bool = True, **kwargs):
        super().__init__(dturns, mean_orbit_rad, bending_rad, trans_gamma,
                         rest_energy, nprofiles, nbins, dtbin, **kwargs)

        kwargs_processed = super()._process_kwargs(_machine_opts_def, kwargs)

        # TODO: Take rfv info as a single input
        # TODO: Take b-field info as a single input

        # Machine parameters
        self.vrf1 = vrf1
        self.vrf1dot = kwargs_processed['vrf1dot']
        self.vrf2 = kwargs_processed['vrf2']
        self.vrf2dot = kwargs_processed['vrf2dot']
        self.b0 = b0
        self.bdot = kwargs_processed['bdot']

        self.fitted_synch_part_x = None
        self.bunchlimit_low = None
        self.bunchlimit_up = None

        if vat_now:
            self.values_at_turns()

    def values_at_turns(self):
        asrt.assert_machine_input(self)
        all_turns = (self.nprofiles - 1) * self.dturns

        # Create all-zero arrays of size nturns+1
        self._init_arrays(all_turns)

        # Calculate initial values at the machine reference frame (i0).
        i0 = self._array_initial_values()

        # Calculate remaining values for every machine turn.
        for i in range(i0 + 1, all_turns + 1):
            self.time_at_turn[i] = (self.time_at_turn[i - 1]
                                    + 2 * np.pi * self.mean_orbit_rad
                                    / (self.beta0[i - 1] * constants.c))

            try:
                self.phi0[i] = optimize.newton(func=physics.rf_voltage_mch,
                                               x0=self.phi0[i - 1],
                                               fprime=physics.drf_voltage_mch,
                                               tol=0.0001,
                                               maxiter=100,
                                               args=(self, i))
            except RuntimeError:
                raise ValueError('Could not fit synchronous phase for the '
                                 'supplied parameters.')

            self.e0[i] = (self.e0[i - 1]
                          + self.q
                          * physics.rf_voltage_at_phase(
                        self.phi0[i], self.vrf1, self.vrf1dot,
                        self.vrf2, self.vrf2dot, self.h_ratio,
                        self.phi12, self.time_at_turn, i))

            self.beta0[i] = np.sqrt(
                1.0 - (self.e_rest / float(self.e0[i])) ** 2)
            self.deltaE0[i] = self.e0[i] - self.e0[i - 1]
        for i in range(i0 - 1, -1, -1):
            self.e0[i] = (self.e0[i + 1]
                          - self.q
                          * physics.rf_voltage_at_phase(
                        self.phi0[i + 1], self.vrf1, self.vrf1dot,
                        self.vrf2, self.vrf2dot, self.h_ratio,
                        self.phi12, self.time_at_turn, i + 1))

            self.beta0[i] = np.sqrt(1.0 - (self.e_rest / self.e0[i]) ** 2)
            self.deltaE0[i] = self.e0[i + 1] - self.e0[i]

            self.time_at_turn[i] = (self.time_at_turn[i + 1]
                                    - 2 * np.pi * self.mean_orbit_rad
                                    / (self.beta0[i] * constants.c))

            try:
                self.phi0[i] = optimize.newton(func=physics.rf_voltage_mch,
                                               x0=self.phi0[i + 1],
                                               fprime=physics.drf_voltage_mch,
                                               tol=0.0001,
                                               maxiter=100,
                                               args=(self, i))
            except RuntimeError:
                raise ValueError('Could not fit synchronous phase for the '
                                 'supplied parameters.')

        # Calculate phase slip factor at each turn
        self.eta0 = physics.phase_slip_factor(self)

        # Calculates dphase for each turn
        self.drift_coef = physics.find_dphase(self)

        # Calculate revolution frequency at each turn
        self.omega_rev0 = physics.revolution_freq(self)

        # Calculate RF-voltages at each turn
        self.vrf1_at_turn, self.vrf2_at_turn = self._rfv_at_turns()

    def load_fitted_synch_part_x_ftn(self,
                                     fit_info: Tuple[float, float, float]):
        """Function for setting the synch_part_x if a fit has been performed.
        Saves parameters retrieved from the fitting routine
        needed by the
        :func:`longitudinal_tomography.compat.fortran.write_plotinfo`
        function in the
        :mod:`longitudinal_tomography.utils.tomo_output`. All needed info
        will be returned from the
        :func:`longitudinal_tomography.data.data_treatment.fit_synch_part_x`
        function.

        Sets the following fields:

        * fitted_synch_part_x
            The new x-coordinate of the synchronous particle (needed for
            :func:`longitudinal_tomography.compat.fortran.write_plotinfo`).
        * bunchlimit_low
            Lower phase of bunch (needed for
            :func:`longitudinal_tomography.compat.fortran.write_plotinfo`).
        * bunchlimit_up
            Upper phase of bunch (needed for
            :func:`longitudinal_tomography.compat.fortran.write_plotinfo`).
        * synch_part_x
            The x-coordinate of the synchronous particle.

        Parameters
        ----------
        fit_info: tuple
            Tuple should hold the following info in the following format:
            (F, L, U), where F is the fitted value of the synchronous particle,
            L is the lower bunch limit, and U is the upper bunch limit. All
            of the values should be given in bins. The info needed by the
            :func:`longitudinal_tomography.compat.fortran.write_plotinfo`
            function if a fit has been performed, and the a Fortran style
            output is to be given during the particle tracking.
        """
        log.info('Saving fitted synch_part_x to machine object.')
        self.fitted_synch_part_x = fit_info[0]
        self.bunchlimit_low = fit_info[1]
        self.bunchlimit_up = fit_info[2]
        self.synch_part_x = self.fitted_synch_part_x

    # Initiating arrays in order to store information about parameters
    # that has a different value every turn.
    def _init_arrays(self, all_turns: int):
        array_length = all_turns + 1
        self.time_at_turn = np.zeros(array_length)
        self.omega_rev0 = np.zeros(array_length)
        self.phi0 = np.zeros(array_length)
        self.drift_coef = np.zeros(array_length)
        self.deltaE0 = np.zeros(array_length)
        self.beta0 = np.zeros(array_length)
        self.eta0 = np.zeros(array_length)
        self.e0 = np.zeros(array_length)

    # Calculating start-values for the parameters that changes for each turn.
    # The reference frame where the start-values
    # are calculated is the machine reference frame.
    def _array_initial_values(self) -> int:
        i0 = self.machine_ref_frame * self.dturns
        self.time_at_turn[i0] = 0
        self.e0[i0] = physics.b_to_e(self)
        self.beta0[i0] = physics.lorentz_beta(self, i0)
        phi_lower, phi_upper = physics.find_phi_lower_upper(self, i0)
        # Synchronous phase of a particle on the nominal orbit
        self.phi0[i0] = physics.find_synch_phase_mch(
            self, i0, phi_lower, phi_upper)
        return i0

    # Using a linear approximation to calculate the RF voltage for each turn.
    def _rfv_at_turns(self) -> Tuple[np.ndarray, np.ndarray]:
        rf1v = self.vrf1 + self.vrf1dot * self.time_at_turn
        rf2v = self.vrf2 + self.vrf2dot * self.time_at_turn
        return rf1v, rf2v



from typing import Any, Optional

import numpy

def back_project(weights: numpy.ndarray[numpy.float64],
                 flat_points: numpy.ndarray[numpy.int32],
                 flat_profiles: numpy.ndarray[numpy.float64],
                 n_particles: int,
                 n_profiles: int) -> numpy.ndarray[numpy.float64]:
    ...

def drift(denergy: numpy.ndarray[numpy.float64],
          dphi: numpy.ndarray[numpy.float64],
          drift_coef: numpy.ndarray[numpy.float64], npart: int, turn: int,
          up: bool = ...) -> numpy.ndarray[numpy.float64]:
    ...

def drift_down(dphi: numpy.ndarray[numpy.float64],
               denergy: numpy.ndarray[numpy.float64], drift_coef: float,
               n_particles: int) -> None:
    ...

def drift_up(dphi: numpy.ndarray[numpy.float64],
             denergy: numpy.ndarray[numpy.float64], drift_coef: float,
             n_particles: int) -> None:
    ...

def kick(machine: object, denergy: numpy.ndarray[numpy.float64],
         dphi: numpy.ndarray[numpy.float64],
         rfv1: numpy.ndarray[numpy.float64],
         rfv2: numpy.ndarray[numpy.float64], npart: int, turn: int,
         up: bool = ...) -> numpy.ndarray[numpy.float64]:
    ...

def kick_and_drift(*args, **kwargs) -> Any:
    ...

def kick_down(dphi: numpy.ndarray[numpy.float64],
              denergy: numpy.ndarray[numpy.float64], rfv1: float, rfv2: float,
              phi0: float, phi12: float, h_ratio: float, n_particles: int,
              acc_kick: float) -> None:
    ...

def kick_up(dphi: numpy.ndarray[numpy.float64],
            denergy: numpy.ndarray[numpy.float64], rfv1: float, rfv2: float,
            phi0: float, phi12: float, h_ratio: float, n_particles: int,
            acc_kick: float) -> None:
    ...

def make_phase_space(xp: numpy.ndarray[numpy.int32],
                     yp: numpy.ndarray[numpy.int32],
                     weights: numpy.ndarray[numpy.float64],
                     n_bins: int) -> numpy.ndarray[numpy.float64]:
    ...

def project(flat_rec: numpy.ndarray[numpy.float64],
            flat_points: numpy.ndarray[numpy.int32],
            weights: numpy.ndarray[numpy.float64], n_particles: int,
            n_profiles: int, n_bins: int) -> numpy.ndarray[numpy.float64]:
    ...

def reconstruct(xp: numpy.ndarray[numpy.int32],
                waterfall: numpy.ndarray[numpy.float64], n_iter: int,
                n_bins: int, n_particles: int, n_profiles: int,
                verbose: bool = ...,
                callback: Optional[object] = ...) -> tuple:
    ...

def reconstruct_multi(xpRound0: numpy.ndarray[numpy.int32],
                      waterfall: numpy.ndarray[numpy.float64],
                      cutleft: numpy.ndarray[numpy.int32],
                      cutright: numpy.ndarray[numpy.int32],
                      centers: numpy.ndarray[numpy.int32],
                      n_iter: int, n_bins: int, n_particles: int,
                      n_profiles: int, n_centers: int,
                      verbose: bool = ...,
                      callback: Optional[object] = ...) -> tuple:
    ...

def set_num_threads(num_threads: int) -> None:
    ...

def make_phase_space(xp: numpy.ndarray[numpy.int32],
                     yp: numpy.ndarray[numpy.int32],
                     weight: numpy.ndarray[numpy.float64],
                     n_particles: int,
                     n_bins: int) -> numpy.ndarray[numpy.float64]:
    ...